import copy
from mindsdb_sql_parser import ast

from mindsdb.api.executor.planner.exceptions import PlanningException
from mindsdb.api.executor.planner import steps
from mindsdb.api.executor.planner import utils


def to_string(identifier):
    # alternative to AST.to_string() but without quoting
    return ".".join(identifier.parts)


class Table:
    def __init__(self, node=None, ds=None, is_predictor=None):
        self.node = node
        self.is_predictor = is_predictor
        self.ds = ds
        self.name = to_string(node)
        self.columns = None
        # None is unknown
        self.columns_map = None
        self.keys = None
        if node.alias:
            self.alias = to_string(node.alias)
        else:
            self.alias = None
            # self.alias = self.name


class Column:
    def __init__(self, node=None, table=None, name=None, type=None):
        alias = None
        if node is not None:
            if isinstance(node, ast.Identifier):
                # set name
                name = node.parts[-1]  # ???

        else:
            if table is not None and name is not None:
                node = ast.Identifier(parts=[table.name, name])

        self.node = node  # link to AST
        self.alias = alias  # for ProjectStep

        self.is_star = False

        self.table = table  # link to AST table
        self.name = name  # column name
        self.type = type


class Statement:
    def __init__(self):
        self.columns = []
        # self.query = None
        self.params = None
        self.result = None

        # Tables on first level of select
        self.tables_lvl1 = None

        # mapping tables by name {'ta.ble': Table()}
        self.tables_map = None

        self.offset = 0


class PreparedStatementPlanner:
    def __init__(self, planner):
        self.planner = planner

    def get_type_of_var(self, v):
        if isinstance(v, str):
            return "str"
        elif isinstance(v, float):
            return "float"
        elif isinstance(v, int):
            return "integer"

        return "str"

    def get_statement_info(self):
        stmt = self.planner.statement

        if stmt is None:
            raise PlanningException("Statement is not prepared")

        columns_result = []

        for column in stmt.columns:
            table, ds = None, None
            if column.table is not None:
                table = column.table.name
                ds = column.table.ds
            columns_result.append(
                dict(
                    alias=column.alias,
                    type=column.type,
                    name=column.name,
                    table_name=table,
                    table_alias=table,
                    ds=ds,
                )
            )

        parameters = []
        for param in stmt.params:
            name = "?"
            parameters.append(
                dict(
                    alias=name,
                    type="str",
                    name=name,
                )
            )

        return {"parameters": parameters, "columns": columns_result}

    def get_table_of_column(self, t):
        tables_map = self.planner.statement.tables_map

        # get tables to check
        if len(t.parts) > 1:
            # try to find table
            table_parts = t.parts[:-1]
            table_name = ".".join(table_parts)
            if table_name in tables_map:
                return tables_map[table_name]

            elif len(table_parts) > 1:
                # maybe datasource is 1st part
                table_parts = table_parts[1:]
                table_name = ".".join(table_parts)
                if table_name in tables_map:
                    return tables_map[table_name]

    def table_from_identifier(self, table):
        # disambiguate
        if self.planner.is_predictor(table):
            ds, table = self.planner.get_predictor_namespace_and_name_from_identifier(table)
            is_predictor = True

        else:
            ds, table = self.planner.resolve_database_table(table)
            is_predictor = False

        if table.alias is not None:
            # access by alias if table is having alias
            keys = [to_string(table.alias)]

        else:
            # access by table name, in all variants
            keys = []
            parts = []
            # in reverse order
            for p in table.parts[::-1]:
                parts.insert(0, p)
                keys.append(".".join(parts))

        # remember table
        tbl = Table(ds=ds, node=table, is_predictor=is_predictor)
        tbl.keys = keys

        return tbl

    def prepare_select(self, query):
        # prepare select with or without predictor

        stmt = self.planner.statement

        # get all predictors
        query_predictors = []

        def find_predictors(node, is_table, **kwargs):
            if is_table and isinstance(node, ast.Identifier):
                if self.planner.is_predictor(node):
                    query_predictors.append(node)

        utils.query_traversal(query, find_predictors)

        # === get all tables from 1st level of query ===
        stmt.tables_map = {}
        stmt.tables_lvl1 = []
        if query.from_table is not None:
            if isinstance(query.from_table, ast.Join):
                # get all tables
                join_tables = utils.convert_join_to_list(query.from_table)
            else:
                join_tables = [dict(table=query.from_table)]

            if isinstance(query.from_table, ast.Select):
                # nested select, get only last select
                join_tables = [dict(table=utils.get_deepest_select(query.from_table).from_table)]

            for i, join_table in enumerate(join_tables):
                table = join_table["table"]
                if isinstance(table, ast.Identifier):
                    tbl = self.table_from_identifier(table)

                    if tbl.is_predictor:
                        # Is the last table?
                        if i + 1 < len(join_tables):
                            raise PlanningException("Predictor must be last table in query")

                    stmt.tables_lvl1.append(tbl)
                    for key in tbl.keys:
                        stmt.tables_map[key] = tbl

                else:
                    # don't add unknown table to looking list
                    continue

        # is there any predictors at other levels?
        lvl1_predictors = [i for i in stmt.tables_lvl1 if i.is_predictor]
        if len(query_predictors) != len(lvl1_predictors):
            raise PlanningException("Predictor is not at first level")

        # === get targets ===
        columns = []
        get_all_tables = False
        for t in query.targets:
            column = Column(t)

            # column alias
            alias = None
            if t.alias is not None:
                alias = to_string(t.alias)

            if isinstance(t, ast.Star):
                if len(stmt.tables_lvl1) == 0:
                    # if "from" is emtpy we can't make plan
                    raise PlanningException("Can't find table")

                column.is_star = True
                get_all_tables = True

            elif isinstance(t, ast.Identifier):
                if alias is None:
                    alias = t.parts[-1]

                table = self.get_table_of_column(t)
                if table is None:
                    # table is not known
                    get_all_tables = True
                else:
                    column.table = table

            elif isinstance(t, ast.Constant):
                if alias is None:
                    alias = str(t.value)
                column.type = self.get_type_of_var(t.value)
            elif isinstance(t, ast.Function):
                # mysql function
                if t.op == "connection_id":
                    column.type = "integer"
                else:
                    column.type = "str"
            else:
                # TODO go down into lower level.
                #  It can be function, operation, select.
                #  But now show it as string

                # TODO add several known types for function, i.e ABS-int

                # TODO TypeCast - as casted type
                column.type = "str"

            if alias is not None:
                column.alias = alias
            columns.append(column)

        # === get columns from tables ===
        request_tables = set()
        for column in columns:
            if column.table is not None:
                request_tables.add(column.table.name)

        for table in stmt.tables_lvl1:
            if get_all_tables or table.name in request_tables:
                if table.is_predictor:
                    step = steps.GetPredictorColumns(namespace=table.ds, predictor=table.node)
                else:
                    step = steps.GetTableColumns(namespace=table.ds, table=table.name)
                yield step

                if step.result_data is not None:
                    # save results

                    if len(step.result_data["tables"]) > 0:
                        table_info = step.result_data["tables"][0]
                        columns_info = step.result_data["columns"][table_info]

                        table.columns = []
                        table.ds = table_info[0]
                        for col in columns_info:
                            if isinstance(col, tuple):
                                # is predictor
                                col = dict(name=col[0], type="str")
                            table.columns.append(
                                Column(
                                    name=col["name"],
                                    type=col["type"],
                                )
                            )

                    # map by names
                    table.columns_map = {i.name.upper(): i for i in table.columns}

        # === create columns list ===
        columns_result = []
        for i, column in enumerate(columns):
            if column.is_star:
                # add data from all tables
                for table in stmt.tables_lvl1:
                    if table.columns is None:
                        raise PlanningException(f"Table is not found {table.name}")

                    for col in table.columns:
                        # col = {name: 'col', type: 'str'}
                        column2 = Column(table=table, name=col.name)
                        column2.alias = col.name
                        column2.type = col.type

                        columns_result.append(column2)

                # to next column
                continue

            elif column.name is not None:
                # is Identifier
                if isinstance(column.name, ast.Star):
                    continue
                col_name = column.name.upper()
                if column.table is not None:
                    table = column.table
                    if table.columns_map is not None:
                        if col_name in table.columns_map:
                            column.type = table.columns_map[col_name].type
                        else:
                            # continue
                            raise PlanningException(f"Column not found {col_name}")

                else:
                    # table is not found, looking for in all tables
                    for table in stmt.tables_lvl1:
                        if table.columns_map is not None:
                            col = table.columns_map.get(col_name)
                            if col is not None:
                                column.type = col.type
                                column.table = table
                                break

            # forcing alias
            if column.alias is None:
                column.alias = f"column_{i}"

            # forcing type
            if column.type is None:
                column.type = "str"

            columns_result.append(column)

        # save columns
        stmt.columns = columns_result

    def prepare_insert(self, query):
        stmt = self.planner.statement

        # get table columns
        table = self.table_from_identifier(query.table)
        if table.is_predictor:
            step = steps.GetPredictorColumns(namespace=table.ds, predictor=table.node)
        else:
            step = steps.GetTableColumns(namespace=table.ds, table=table.name)
        yield step

        if step.result_data is not None:
            # save results

            if len(step.result_data["tables"]) > 0:
                table_info = step.result_data["tables"][0]
                columns_info = step.result_data["columns"][table_info]

                table.columns = []
                table.ds = table_info[0]
                for col in columns_info:
                    if isinstance(col, tuple):
                        # is predictor
                        col = dict(name=col[0], type="str")
                    table.columns.append(
                        Column(
                            name=col["name"],
                            type=col["type"],
                        )
                    )

                # map by names
                table.columns_map = {i.name.upper(): i for i in table.columns}

        # save results
        columns_result = []
        for col in query.columns:
            col_name = col.parts[-1]

            column = Column(table=table, name=col_name)

            if table.columns_map is not None:
                col = table.columns_map.get(col_name)
                if col is not None:
                    column.type = col.type

            if column.type is None:
                # forcing type
                column.type = "str"

            columns_result.append(column)

        stmt.columns = columns_result

    def prepare_show(self, query):
        stmt = self.planner.statement

        stmt.columns = [
            Column(name="Variable_name", type="str"),
            Column(name="Value", type="str"),
        ]
        return []

    def prepare_steps(self, query):
        stmt = Statement()
        self.planner.statement = stmt

        self.planner.query = query

        query = copy.deepcopy(query)

        params = utils.get_query_params(query)

        stmt.params = params

        # get columns
        if isinstance(query, ast.Select):
            # prepare select
            return self.prepare_select(query)
        if isinstance(query, ast.Union):
            # get column definition only from select
            return self.prepare_select(query.left)
        if isinstance(query, ast.Insert):
            # return self.prepare_insert(query)
            # TODO do we need columns?
            return []
        if isinstance(query, ast.Delete):
            ...
            # TODO do we need columns?
            return []
        if isinstance(query, ast.Show):
            return self.prepare_show(query)
        else:
            # do nothing
            return []
            # raise NotImplementedError(query.__name__)

    def execute_steps(self, params=None):
        # find all parameters
        stmt = self.planner.statement

        # is already executed
        if stmt is None:
            if params is not None:
                raise PlanningException("Can't execute statement")
            stmt = Statement()

        # === form query with new target ===

        query = self.planner.query

        if params is not None:
            if len(params) != len(stmt.params):
                raise PlanningException("Count of execution parameters don't match prepared statement")

            query = utils.fill_query_params(query, params)

            self.planner.query = query

        # prevent from second execution
        stmt.params = None

        if (
            isinstance(query, ast.Select)
            or isinstance(query, ast.Union)
            or isinstance(query, ast.CreateTable)
            or isinstance(query, ast.Insert)
            or isinstance(query, ast.Update)
            or isinstance(query, ast.Delete)
            or isinstance(query, ast.Intersect)
            or isinstance(query, ast.Except)
        ):
            return self.plan_query(query)
        else:
            return []

    def plan_query(self, query):
        # use v1 planner
        self.planner.from_query(query)
        step = None
        for step in self.planner.plan.steps:
            yield step



from mindsdb_sql_parser.ast import Select
from typing import List
import pandas as pd

from mindsdb.integrations.utilities.sql_utils import FilterCondition, KeywordSearchArgs


class KeywordSearchBase:
    """
    Base class for keyword search integrations.
    This class provides a common interface for keyword search functionality.
    """

    def __init__(self, *args, **kwargs):
        pass

    def dispatch_keyword_select(
        self, query: Select, conditions: List[FilterCondition] = None, keyword_search_args: KeywordSearchArgs = None
    ):
        """Dispatches a keyword search select query to the appropriate method."""
        raise NotImplementedError()

    def keyword_select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
    ) -> pd.DataFrame:
        """Select data from table

        Args:
            table_name (str): table name
            columns (List[str]): columns to select
            conditions (List[FilterCondition]): conditions to select

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()



from pydantic_settings import BaseSettings


class TogetherAIConfig(BaseSettings):
    """
    Configuration for TogetherAI handler.

    Attributes
    ----------

    BASE_URL : str
        Base URL for the TogetherAI API.
    DEFAULT_MODEL : str
        Default model to use for TogetherAI API.
    DEFAULT_MODE : str
        Defat mode to use for TogetherAI API.
    SUPPORTED_MODES : list[str]
        List of supported modes for TogetherAI API.
    """

    BASE_URL: str = "https://api.together.xyz/v1"
    DEFAULT_MODEL: str = "meta-llama/Llama-3.2-3B-Instruct-Turbo"
    DEFAULT_EMBEDDING_MODEL: str = "togethercomputer/m2-bert-80M-32k-retrieval"
    DEFAULT_MODE: str = "default"
    SUPPORTED_MODES: list[str] = [
        "default",
        "conversational",
        "conversational-full",
        "embedding",
    ]


togetherai_handler_config = TogetherAIConfig()



import unittest
from mindsdb.integrations.handlers.databend_handler.databend_handler import DatabendHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class DatabendHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        connection_data = {
            "host": "some-url.aws-us-east-2.default.databend.com",
            "port": 443,
            "user": "root",
            "password": "password",
            "database": "test_db"
        }
        cls.handler = DatabendHandler('test_databend_handler', connection_data)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_select_query(self):
        query = 'SELECT * FROM covid_19_us_2022_4668 LIMIT 10'
        result = self.handler.query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_2_get_tables(self):
        tbls = self.handler.get_tables()
        assert tbls.type is not RESPONSE_TYPE.ERROR

    def test_3_describe_table(self):
        described = self.handler.get_columns("covid_19_us_2022_4668")
        print('described', described)
        assert described.type is RESPONSE_TYPE.TABLE


if __name__ == '__main__':
    unittest.main()



import json
import time
from typing import AsyncIterable, Any, Dict

from starlette.applications import Starlette
from starlette.middleware.cors import CORSMiddleware
from starlette.responses import JSONResponse
from sse_starlette.sse import EventSourceResponse
from starlette.requests import Request
from starlette.routing import Route
from ...common.types import (
    A2ARequest,
    JSONRPCResponse,
    InvalidRequestError,
    JSONParseError,
    GetTaskRequest,
    CancelTaskRequest,
    SendTaskRequest,
    SetTaskPushNotificationRequest,
    GetTaskPushNotificationRequest,
    InternalError,
    AgentCard,
    TaskResubscriptionRequest,
    SendTaskStreamingRequest,
)
from pydantic import ValidationError
from ...common.server.task_manager import TaskManager

from mindsdb.utilities import log

logger = log.getLogger(__name__)


class A2AServer:
    def __init__(
        self,
        agent_card: AgentCard = None,
        task_manager: TaskManager = None,
    ):
        self.task_manager = task_manager
        self.agent_card = agent_card
        self.app = Starlette(
            routes=[
                Route("/", self._process_request, methods=["POST"]),
                Route("/.well-known/agent.json", self._get_agent_card, methods=["GET"]),
                Route("/status", self._get_status, methods=["GET"]),
            ]
        )
        # TODO: Remove this when we have a proper CORS policy
        self.app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
        self.start_time = time.time()

    def _get_agent_card(self, request: Request) -> JSONResponse:
        return JSONResponse(self.agent_card.model_dump(exclude_none=True))

    def _get_status(self, request: Request) -> JSONResponse:
        """
        Status endpoint that returns basic server information.
        This endpoint can be used by the frontend to check if the A2A server is running.
        """
        uptime_seconds = time.time() - self.start_time

        status_info: Dict[str, Any] = {
            "status": "ok",
            "service": "mindsdb-a2a",
            "uptime_seconds": round(uptime_seconds, 2),
            "agent_name": self.agent_card.name if self.agent_card else None,
            "version": self.agent_card.version if self.agent_card else "unknown",
        }

        return JSONResponse(status_info)

    async def _process_request(self, request: Request):
        try:
            body = await request.json()
            json_rpc_request = A2ARequest.validate_python(body)

            user_info = {
                "user-id": request.headers.get("user-id", None),
                "company-id": request.headers.get("company-id", None),
                "user-class": request.headers.get("user-class", None),
                "authorization": request.headers.get("Authorization", None),
            }

            if isinstance(json_rpc_request, GetTaskRequest):
                result = await self.task_manager.on_get_task(json_rpc_request)
            elif isinstance(json_rpc_request, SendTaskRequest):
                result = await self.task_manager.on_send_task(json_rpc_request, user_info)
            elif isinstance(json_rpc_request, SendTaskStreamingRequest):
                # Don't await the async generator, just pass it to _create_response
                result = self.task_manager.on_send_task_subscribe(json_rpc_request, user_info)
            elif isinstance(json_rpc_request, CancelTaskRequest):
                result = await self.task_manager.on_cancel_task(json_rpc_request)
            elif isinstance(json_rpc_request, SetTaskPushNotificationRequest):
                result = await self.task_manager.on_set_task_push_notification(json_rpc_request)
            elif isinstance(json_rpc_request, GetTaskPushNotificationRequest):
                result = await self.task_manager.on_get_task_push_notification(json_rpc_request)
            elif isinstance(json_rpc_request, TaskResubscriptionRequest):
                result = await self.task_manager.on_resubscribe_to_task(json_rpc_request)
            else:
                logger.warning(f"Unexpected request type: {type(json_rpc_request)}")
                raise ValueError(f"Unexpected request type: {type(request)}")

            return self._create_response(result)

        except Exception as e:
            return self._handle_exception(e)

    def _handle_exception(self, e: Exception) -> JSONResponse:
        if isinstance(e, json.decoder.JSONDecodeError):
            json_rpc_error = JSONParseError()
        elif isinstance(e, ValidationError):
            json_rpc_error = InvalidRequestError(data=json.loads(e.json()))
        else:
            logger.error(f"Unhandled exception: {e}")
            json_rpc_error = InternalError()

        response = JSONRPCResponse(id=None, error=json_rpc_error)
        return JSONResponse(response.model_dump(exclude_none=True), status_code=400)

    def _create_response(self, result: Any) -> JSONResponse | EventSourceResponse:
        if isinstance(result, AsyncIterable):
            # Step 2: Yield actual serialized event as JSON, with timing logs
            async def event_generator(result):
                async for item in result:
                    t0 = time.time()
                    logger.debug(f"[A2AServer] STEP2 serializing item at {t0}: {str(item)[:120]}")
                    try:
                        if hasattr(item, "model_dump_json"):
                            data = item.model_dump_json(exclude_none=True)
                        else:
                            data = json.dumps(item)
                    except Exception as e:
                        logger.error(f"Serialization error in SSE stream: {e}")
                        data = json.dumps({"error": f"Serialization error: {str(e)}"})
                    yield {"data": data}

            # Add robust SSE headers for compatibility
            sse_headers = {
                "Content-Type": "text/event-stream",
                "Cache-Control": "no-cache, no-transform",
                "X-Accel-Buffering": "no",
                "Connection": "keep-alive",
                "Transfer-Encoding": "chunked",
            }
            return EventSourceResponse(event_generator(result), headers=sse_headers)
        elif isinstance(result, JSONRPCResponse):
            return JSONResponse(result.model_dump(exclude_none=True))
        elif isinstance(result, dict):
            logger.warning("Falling back to JSONResponse for result type: dict")
            return JSONResponse(result)
        else:
            logger.error(f"Unexpected result type: {type(result)}")
            raise ValueError(f"Unexpected result type: {type(result)}")



from typing import List
from datetime import datetime, timedelta
import pandas as pd

from mindsdb.integrations.libs.api_handler import APIResource
from mindsdb.integrations.utilities.sql_utils import FilterCondition, FilterOperator, SortColumn
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class GongCallsTable(APIResource):
    """The Gong Calls Table implementation"""

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
    ) -> pd.DataFrame:
        """Pulls data from the Gong Calls API

        Returns
        -------
        pd.DataFrame
            Gong calls matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        api_params = {}
        if conditions:
            for condition in conditions:
                if condition.column == "date" and condition.op == FilterOperator.GREATER_THAN:
                    api_params["fromDateTime"] = condition.value
                    condition.applied = True
                elif condition.column == "date" and condition.op == FilterOperator.LESS_THAN:
                    api_params["toDateTime"] = condition.value
                    condition.applied = True

        try:
            all_calls = []
            cursor = None

            while True:
                current_params = api_params.copy()
                if cursor:
                    current_params["cursor"] = cursor

                response = self.handler.call_gong_api("/v2/calls", current_params)
                calls_batch = response.get("calls", [])

                for call in calls_batch:
                    if limit and len(all_calls) >= limit:
                        break
                    all_calls.append(call)

                records_info = response.get("records", {})
                if (limit and len(all_calls) >= limit) or "cursor" not in records_info:
                    break

                cursor = records_info.get("cursor")
                if not cursor:
                    break

            # Process the limited data
            data = []
            for call in all_calls:
                item = {
                    "call_id": call.get("id"),
                    "title": call.get("title"),
                    "date": call.get("started").split("T")[0],
                    "duration": call.get("duration"),
                    "recording_url": call.get("url", ""),
                    "call_type": call.get("system"),
                    "user_id": call.get("primaryUserId"),
                    "participants": ",".join([p.get("name", "") for p in call.get("participants", [])]),
                    "status": call.get("status"),
                }
                data.append(item)

            df = pd.DataFrame(data)

            # Apply non-date filtering at DataFrame level
            if conditions:
                for condition in conditions:
                    if not condition.applied and condition.column in df.columns:
                        if condition.op == FilterOperator.EQUAL:
                            df = df[df[condition.column] == condition.value]
                            condition.applied = True

            if sort:
                for col in sort:
                    if col.column in df.columns:
                        df = df.sort_values(by=col.column, ascending=col.ascending, na_position="last")
                        col.applied = True
                        break

            if limit is not None:
                df = df.head(limit)

            return df

        except Exception as e:
            logger.error(f"Error fetching calls from Gong API: {e}")
            raise

    def get_columns(self) -> List[str]:
        """Returns the columns of the calls table"""
        return [
            "call_id",
            "title",
            "date",
            "duration",
            "recording_url",
            "call_type",
            "user_id",
            "participants",
            "status",
        ]


class GongUsersTable(APIResource):
    """The Gong Users Table implementation"""

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
    ) -> pd.DataFrame:
        """Pulls data from the Gong Users API

        Returns
        -------
        pd.DataFrame
            Gong users matching the query
        """

        api_params = {}

        try:
            all_users = []
            cursor = None

            while True:
                current_params = api_params.copy()
                if cursor:
                    current_params["cursor"] = cursor

                response = self.handler.call_gong_api("/v2/users", current_params)
                users_batch = response.get("users", [])

                for user in users_batch:
                    if limit and len(all_users) >= limit:
                        break
                    all_users.append(user)

                records_info = response.get("records", {})
                if (limit and len(all_users) >= limit) or "cursor" not in records_info:
                    break

                cursor = records_info.get("cursor")
                if not cursor:
                    break

            # Process the limited data
            data = []
            for user in all_users:
                item = {
                    "user_id": user.get("id"),
                    "name": user.get("firstName") + " " + user.get("lastName"),
                    "email": user.get("emailAddress"),
                    "role": user.get("title"),
                    "permissions": ",".join(user.get("permissions", [])),
                    "status": "active" if user.get("active") else "inactive",
                }
                data.append(item)

            df = pd.DataFrame(data)

            if conditions:
                for condition in conditions:
                    if condition.column in df.columns:
                        if condition.op == FilterOperator.EQUAL:
                            df = df[df[condition.column] == condition.value]
                            condition.applied = True

            if sort:
                for col in sort:
                    if col.column in df.columns:
                        df = df.sort_values(by=col.column, ascending=col.ascending, na_position="last")
                        col.applied = True
                        break

            if limit is not None:
                df = df.head(limit)

            return df

        except Exception as e:
            logger.error(f"Error fetching users from Gong API: {e}")
            raise

    def get_columns(self) -> List[str]:
        """Returns the columns of the users table"""
        return ["user_id", "name", "email", "role", "permissions", "status"]


class GongAnalyticsTable(APIResource):
    """The Gong Analytics Table implementation"""

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
    ) -> pd.DataFrame:
        """Pulls data from the Gong Analytics API

        Returns
        -------
        pd.DataFrame
            Gong analytics matching the query
        """

        try:
            payload = {
                "filter": {
                    "fromDateTime": (datetime.now() - timedelta(days=7)).strftime("%Y-%m-%dT%H:%M:%SZ"),
                    "toDateTime": datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ"),
                },
                "contentSelector": {
                    "exposedFields": {
                        "content": {
                            "brief": True,
                            "outline": True,
                            "highlights": True,
                            "callOutcome": True,
                            "topics": True,
                            "trackers": True,
                        },
                        "interaction": {"personInteractionStats": True, "questions": True},
                    }
                },
            }

            if conditions:
                for condition in conditions:
                    if condition.column == "date" and condition.op == FilterOperator.GREATER_THAN:
                        payload["filter"]["fromDateTime"] = condition.value
                        condition.applied = True
                    elif condition.column == "date" and condition.op == FilterOperator.LESS_THAN:
                        payload["filter"]["toDateTime"] = condition.value
                        condition.applied = True

            session = self.handler.connect()

            all_analytics = []
            cursor = None

            while True:
                current_payload = payload.copy()
                if cursor:
                    current_payload["cursor"] = cursor

                response = session.post(f"{self.handler.base_url}/v2/calls/extensive", json=current_payload)
                response.raise_for_status()
                calls_response = response.json()

                analytics_batch = calls_response.get("calls", [])

                # Process and add analytics
                for call in analytics_batch:
                    if limit and len(all_analytics) >= limit:
                        break

                    # Extract analytics from extensive call data
                    content = call.get("content", {})
                    interaction = call.get("interaction", {})
                    metadata = call.get("metaData", {})

                    # Sentiment and Emotion from InteractionStats
                    person_stats = interaction.get("interactionStats", [])
                    sentiment_score = 0
                    if person_stats:
                        stats_dict = {stat["name"]: stat["value"] for stat in interaction.get("interactionStats", [])}
                        sentiment_score = (
                            stats_dict.get("Talk Ratio", 0)
                            + stats_dict.get("Patience", 0)
                            + min(stats_dict.get("Interactivity", 0) / 10, 1.0)
                        ) / 3
                        emotions = f"Talk:{stats_dict.get('Talk Ratio', 0)}, Patience:{stats_dict.get('Patience', 0)}, Interactivity:{stats_dict.get('Interactivity', 0)}"

                    # Topics from AI analysis
                    topics = content.get("topics", [])
                    topic_names = [
                        topic.get("name", "")
                        for topic in topics
                        if isinstance(topic, dict) and topic.get("duration", 0) > 0
                    ]

                    # Key phrases from AI
                    trackers = content.get("trackers", [])
                    key_phrases = [tracker.get("name", "") for tracker in trackers if tracker.get("count", 0) > 0]

                    # Topic scoring based on relevance
                    topic_duration = (
                        sum([topic.get("duration", 0) for topic in topics if isinstance(topic, dict)]) / len(topics)
                        if topics
                        else 0
                    )
                    call_duration = metadata.get("duration", 1)
                    topic_score = (topic_duration / call_duration) if call_duration > 0 else 0

                    item = {
                        "call_id": metadata.get("id"),
                        "sentiment_score": round(sentiment_score, 3),
                        "topic_score": round(topic_score, 3),
                        "key_phrases": ", ".join(key_phrases),
                        "topics": ", ".join(topic_names),
                        "emotions": emotions,
                        "confidence_score": "",
                    }
                    all_analytics.append(item)

                records_info = calls_response.get("records", {})
                if (limit and len(all_analytics) >= limit) or "cursor" not in records_info:
                    break

                cursor = records_info.get("cursor")
                if not cursor:
                    break

            df = pd.DataFrame(all_analytics)

            # Apply non-date filtering at DataFrame level
            if conditions:
                for condition in conditions:
                    if not condition.applied and condition.column in df.columns:
                        if condition.op == FilterOperator.EQUAL:
                            df = df[df[condition.column] == condition.value]
                            condition.applied = True

            # Apply sorting at DataFrame level
            if sort:
                for col in sort:
                    if col.column in df.columns:
                        df = df.sort_values(by=col.column, ascending=col.ascending, na_position="last")
                        col.applied = True
                        break

            if limit is not None:
                df = df.head(limit)

            return df

        except Exception as e:
            logger.error(f"Error fetching analytics from Gong API: {e}")
            raise

    def get_columns(self) -> List[str]:
        """Returns the columns of the analytics table"""
        return ["call_id", "sentiment_score", "topic_score", "key_phrases", "topics", "emotions", "confidence_score"]


class GongTranscriptsTable(APIResource):
    """The Gong Transcripts Table implementation"""

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
    ) -> pd.DataFrame:
        """Pulls data from the Gong Transcripts API

        Returns
        -------
        pd.DataFrame
            Gong transcripts matching the query
        """

        try:
            calls_api_params = {}

            if conditions:
                for condition in conditions:
                    if condition.column == "date" and condition.op == FilterOperator.GREATER_THAN:
                        calls_api_params["fromDateTime"] = condition.value
                        condition.applied = True
                    elif condition.column == "date" and condition.op == FilterOperator.LESS_THAN:
                        calls_api_params["toDateTime"] = condition.value
                        condition.applied = True

            calls_fetch_limit = limit if limit else 100

            all_call_ids = []
            cursor = None

            while len(all_call_ids) < calls_fetch_limit:
                current_params = calls_api_params.copy()
                if cursor:
                    current_params["cursor"] = cursor

                calls_response = self.handler.call_gong_api("/v2/calls", current_params)
                calls_batch = calls_response.get("calls", [])

                batch_call_ids = [call.get("id") for call in calls_batch if call.get("id")]
                all_call_ids.extend(batch_call_ids)

                records_info = calls_response.get("records", {})
                if len(all_call_ids) >= calls_fetch_limit or "cursor" not in records_info:
                    break

                cursor = records_info.get("cursor")
                if not cursor:
                    break

            if not all_call_ids:
                return pd.DataFrame()
            call_ids_to_fetch = all_call_ids[:limit] if limit else all_call_ids

            session = self.handler.connect()
            all_transcript_data = []
            transcript_cursor = None

            while True:
                payload = {"filter": {"callIds": call_ids_to_fetch}}

                if transcript_cursor:
                    payload["cursor"] = transcript_cursor

                response = session.post(f"{self.handler.base_url}/v2/calls/transcript", json=payload)
                response.raise_for_status()
                transcript_response = response.json()
                transcript_batch = transcript_response.get("callTranscripts", [])

                for call_transcript in transcript_batch:
                    call_id = call_transcript.get("callId")
                    transcript_segments = call_transcript.get("transcript", [])

                    segment_counter = 0

                    for speaker_block in transcript_segments:
                        speaker_id = speaker_block.get("speakerId")
                        sentences = speaker_block.get("sentences", [])

                        for sentence in sentences:
                            segment_counter += 1

                            item = {
                                "call_id": call_id,
                                "speaker": speaker_id,
                                "timestamp": sentence.get("start"),
                                "text": sentence.get("text"),
                                "confidence": sentence.get("confidence"),
                                "segment_id": f"{call_id}_{segment_counter}",
                            }
                            all_transcript_data.append(item)

                transcript_records_info = transcript_response.get("records", {})
                if "cursor" not in transcript_records_info:
                    break

                transcript_cursor = transcript_records_info.get("cursor")
                if not transcript_cursor:
                    break

            df = pd.DataFrame(all_transcript_data)

            if conditions:
                for condition in conditions:
                    if not condition.applied and condition.column in df.columns:
                        if condition.op == FilterOperator.EQUAL:
                            df = df[df[condition.column] == condition.value]
                            condition.applied = True
                        elif condition.op == FilterOperator.LIKE or condition.op == FilterOperator.CONTAINS:
                            if condition.column == "text":
                                df = df[df[condition.column].str.contains(condition.value, case=False, na=False)]
                                condition.applied = True

            if sort:
                for col in sort:
                    if col.column in df.columns:
                        df = df.sort_values(by=col.column, ascending=col.ascending, na_position="last")
                        col.applied = True
                        break

            if limit is not None:
                df = df.head(limit)
            return df

        except Exception as e:
            logger.error(f"Error fetching transcripts from Gong API: {e}")
            raise

    def get_columns(self) -> List[str]:
        """Returns the columns of the transcripts table"""
        return ["call_id", "speaker", "timestamp", "text", "confidence", "segment_id"]



from pymongo import MongoClient

from mindsdb.integrations.handlers.mongodb_handler import Handler as MongoDBHandler


class DocumentDBHandler(MongoDBHandler):
    """
    This handler handles connection and execution of the DocumentDB statements.
    """

    name = 'documentdb'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        connection_data = kwargs.get('connection_data', {})
        self.host = connection_data.get("host")
        self.port = int(connection_data.get("port") or 27017)
        self.user = connection_data.get("username")
        self.password = connection_data.get("password")
        self.database = connection_data.get('database')
        self.flatten_level = connection_data.get('flatten_level', 0)
        self.mykwargs = connection_data.get('kwargs', {})

        self.connection = None
        self.is_connected = False

    def connect(self):
        kwargs = {}
        if isinstance(self.user, str) and len(self.user) > 0:
            kwargs['username'] = self.user

        if isinstance(self.password, str) and len(self.password) > 0:
            kwargs['password'] = self.password

        connection = MongoClient(
            host=self.host, port=self.port, **{**kwargs, **self.mykwargs}
        )

        self.is_connected = True
        self.connection = connection
        return self.connection



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'Rockset user name'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'Rockset password',
        'secret': True
    },
    api_key={
        'type': ARG_TYPE.STR,
        'description': 'Rockset API key'
    },
    api_server={
        'type': ARG_TYPE.STR,
        'description': 'Rockset API server'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'Rockset host'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Rockset port'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'Rockset database'
    }
)
connection_args_example = OrderedDict(
    user='rockset',
    password='rockset',
    api_key="adkjf234rksjfa23waejf2",
    api_server='api-us-west-2.rockset.io',
    host='localhost',
    port='3306',
    database='test'
)



"""query_database

Revision ID: 53502b6d63bf
Revises: fda503400e43
Create Date: 2025-04-22 16:30:15.139978

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa


# revision identifiers, used by Alembic.
revision = '53502b6d63bf'
down_revision = 'fda503400e43'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('queries', schema=None) as batch_op:
        batch_op.add_column(sa.Column('database', sa.String(), nullable=True))


def downgrade():
    with op.batch_alter_table('queries', schema=None) as batch_op:
        batch_op.drop_column('database')



class TableNotFound(Exception):
    pass


class ConnectionFailed(Exception):
    pass


class InvalidNativeQuery(Exception):
    pass


class TableAlreadyExists(Exception):
    pass


class MissingConnectionParams(Exception):
    pass



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    username={
        'type': ARG_TYPE.STR,
        'description': 'The username for the Salesforce account.',
        'required': True,
        'label': 'Username'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password for the Salesforce account.',
        'secret': True,
        'required': True,
        'label': 'Password'
    },
    client_id={
        'type': ARG_TYPE.STR,
        'description': 'The client ID (consumer key) from a connected app in Salesforce.',
        'required': True,
        'label': 'Client ID (Consumer Key)'
    },
    client_secret={
        'type': ARG_TYPE.STR,
        'description': 'The client secret (consumer secret) from a connected app in Salesforce.',
        'required': True,
        'label': 'Client Secret (Consumer Secret)'
    },
    is_sandbox={
        'type': ARG_TYPE.BOOL,
        'description': 'Set this to True if you need to connect to a sandbox, False for production environments. '
                       'If not provided defaults to False.',
        'required': False,
        'label': 'Is Sandbox'
    }
)

connection_args_example = OrderedDict(
    username='demo@example.com',
    password='demo_password',
    client_id='3MVG9lKcPoNINVBIPJjdw1J9LLM82HnZz9Yh7ZJnY',
    client_secret='5A52C1A1E21DF9012IODC9ISNXXAADDA9',
    is_sandbox=True
)



from typing import Dict, List
from mindsdb.utilities.log import getLogger

logger = getLogger(__name__)


def to_serializable(obj):
    # Primitives
    if isinstance(obj, (str, int, float, bool, type(None))):
        return obj
    # Pydantic v2
    if hasattr(obj, "model_dump"):
        return to_serializable(obj.model_dump(exclude_none=True))
    # Pydantic v1
    if hasattr(obj, "dict"):
        return to_serializable(obj.dict(exclude_none=True))
    # Custom classes with __dict__
    if hasattr(obj, "__dict__"):
        return {k: to_serializable(v) for k, v in vars(obj).items() if not k.startswith("_")}
    # Dicts
    if isinstance(obj, dict):
        return {k: to_serializable(v) for k, v in obj.items()}
    # Lists, Tuples, Sets
    if isinstance(obj, (list, tuple, set)):
        return [to_serializable(v) for v in obj]
    # Fallback: string
    return str(obj)


def convert_a2a_message_to_qa_format(a2a_message: Dict) -> List[Dict[str, str]]:
    """
    Convert A2A message format to question/answer format.

    This is the format that the langchain agent expects and ensure effective multi-turn conversation

    Args:
        a2a_message: A2A message containing history and current message parts

    Returns:
        List of messages in question/answer format
    """
    converted_messages = []

    # Process conversation history first
    if "history" in a2a_message and a2a_message["history"] is not None:
        for hist_msg in a2a_message["history"]:
            if hist_msg.get("role") == "user":
                # Extract text from parts
                text = ""
                for part in hist_msg.get("parts", []):
                    if part.get("type") == "text":
                        text = part.get("text", "")
                        break
                # Create question with empty answer initially
                converted_messages.append({"question": text, "answer": ""})
            elif hist_msg.get("role") in ["agent", "assistant"]:
                # Extract text from parts
                text = ""
                for part in hist_msg.get("parts", []):
                    if part.get("type") == "text":
                        text = part.get("text", "")
                        break
                # Pair with the most recent question that has empty answer
                paired = False
                for i in range(len(converted_messages) - 1, -1, -1):
                    if converted_messages[i].get("answer") == "":
                        converted_messages[i]["answer"] = text
                        paired = True
                        break

                if not paired:
                    logger.warning("Could not pair agent response with question (no empty answer found)")

        logger.debug(f"Converted {len(a2a_message['history'])} A2A history messages to Q&A format")

    # Add current message as final question with empty answer
    current_text = ""
    for part in a2a_message.get("parts", []):
        if part.get("type") == "text":
            current_text = part.get("text", "")
            break
    converted_messages.append({"question": current_text, "answer": ""})

    return converted_messages



from dataclasses import dataclass, field
from typing import List, Dict

import pandas as pd

from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import MYSQL_DATA_TYPE


@dataclass
class DataHubResponse:
    data_frame: pd.DataFrame = field(default_factory=pd.DataFrame)
    columns: List[Dict] = field(default_factory=list)
    affected_rows: int | None = None
    mysql_types: list[MYSQL_DATA_TYPE] | None = None



import pandas as pd
import mysql.connector

from mindsdb_sql_parser import parse_sql
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)
from mindsdb.integrations.handlers.mysql_handler.settings import ConnectionConfig
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import MYSQL_DATA_TYPE
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import C_TYPES, DATA_C_TYPE_MAP

logger = log.getLogger(__name__)


def _map_type(mysql_type_text: str) -> MYSQL_DATA_TYPE:
    """Map MySQL text types names to MySQL types as enum.

    Args:
        mysql_type_text (str): The name of the MySQL type to map.

    Returns:
        MYSQL_DATA_TYPE: The MySQL type enum that corresponds to the MySQL text type name.
    """
    try:
        return MYSQL_DATA_TYPE(mysql_type_text.upper())
    except Exception:
        logger.warning(f"MySQL handler: unknown type: {mysql_type_text}, use TEXT as fallback.")
        return MYSQL_DATA_TYPE.TEXT


def _make_table_response(result: list[dict], cursor: mysql.connector.cursor.MySQLCursor) -> Response:
    """Build response from result and cursor.

    Args:
        result (list[dict]): result of the query.
        cursor (mysql.connector.cursor.MySQLCursor): cursor object.

    Returns:
        Response: response object.
    """
    description = cursor.description
    reverse_c_type_map = {v.code: k for k, v in DATA_C_TYPE_MAP.items() if v.code != C_TYPES.MYSQL_TYPE_BLOB}
    mysql_types: list[MYSQL_DATA_TYPE] = []
    for col in description:
        type_int = col[1]
        if isinstance(type_int, int) is False:
            mysql_types.append(MYSQL_DATA_TYPE.TEXT)
            continue

        if type_int == C_TYPES.MYSQL_TYPE_TINY:
            # There are 3 types that returns as TINYINT: TINYINT, BOOL, BOOLEAN.
            mysql_types.append(MYSQL_DATA_TYPE.TINYINT)
            continue

        if type_int in reverse_c_type_map:
            mysql_types.append(reverse_c_type_map[type_int])
            continue

        if type_int == C_TYPES.MYSQL_TYPE_BLOB:
            # region determine text/blob type by flags
            # Unfortunately, there is no way to determine particular type of text/blob column by flags.
            # Subtype have to be determined by 8-s element of description tuple, but mysql.conector
            # return the same value for all text types (TINYTEXT, TEXT, MEDIUMTEXT, LONGTEXT), and for
            # all blob types (TINYBLOB, BLOB, MEDIUMBLOB, LONGBLOB).
            if col[7] == 16:  # and col[8] == 45
                mysql_types.append(MYSQL_DATA_TYPE.TEXT)
            elif col[7] == 144:  # and col[8] == 63
                mysql_types.append(MYSQL_DATA_TYPE.BLOB)
            else:
                logger.debug(f"MySQL handler: unknown type code {col[7]}, use TEXT as fallback.")
                mysql_types.append(MYSQL_DATA_TYPE.TEXT)
            # endregion
        else:
            logger.warning(f"MySQL handler: unknown type id={type_int} in column {col[0]}, use TEXT as fallback.")
            mysql_types.append(MYSQL_DATA_TYPE.TEXT)

    # region cast int and bool to nullable types
    serieses = []
    for i, mysql_type in enumerate(mysql_types):
        expected_dtype = None
        column_name = description[i][0]
        if mysql_type in (
            MYSQL_DATA_TYPE.SMALLINT,
            MYSQL_DATA_TYPE.INT,
            MYSQL_DATA_TYPE.MEDIUMINT,
            MYSQL_DATA_TYPE.BIGINT,
            MYSQL_DATA_TYPE.TINYINT,
        ):
            expected_dtype = "Int64"
        elif mysql_type in (MYSQL_DATA_TYPE.BOOL, MYSQL_DATA_TYPE.BOOLEAN):
            expected_dtype = "boolean"
        serieses.append(pd.Series([row[column_name] for row in result], dtype=expected_dtype, name=description[i][0]))
    df = pd.concat(serieses, axis=1, copy=False)
    # endregion

    response = Response(RESPONSE_TYPE.TABLE, df, affected_rows=cursor.rowcount, mysql_types=mysql_types)
    return response


class MySQLHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the MySQL statements.
    """

    name = "mysql"

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = "mysql"
        self.connection_data = kwargs.get("connection_data", {})
        self.database = self.connection_data.get("database")

        self.connection = None

    def __del__(self):
        if self.is_connected:
            self.disconnect()

    def _unpack_config(self):
        """
        Unpacks the config from the connection_data by validation all parameters.

        Returns:
            dict: A dictionary containing the validated connection parameters.
        """
        try:
            config = ConnectionConfig(**self.connection_data)
            return config.model_dump(exclude_unset=True)
        except ValueError as e:
            raise ValueError(str(e))

    @property
    def is_connected(self):
        """
        Checks if the handler is connected to the MySQL database.

        Returns:
            bool: True if the handler is connected, False otherwise.
        """
        return self.connection is not None and self.connection.is_connected()

    @is_connected.setter
    def is_connected(self, value):
        pass

    def connect(self):
        """
        Establishes a connection to a MySQL database.

        Returns:
            MySQLConnection: An active connection to the database.
        """
        if self.is_connected and self.connection.is_connected():
            return self.connection
        config = self._unpack_config()
        if "conn_attrs" in self.connection_data:
            config["conn_attrs"] = self.connection_data["conn_attrs"]

        if "connection_timeout" not in config:
            config["connection_timeout"] = 10

        ssl = self.connection_data.get("ssl")
        if ssl is True:
            ssl_ca = self.connection_data.get("ssl_ca")
            ssl_cert = self.connection_data.get("ssl_cert")
            ssl_key = self.connection_data.get("ssl_key")
            config["client_flags"] = [mysql.connector.constants.ClientFlag.SSL]
            if ssl_ca is not None:
                config["ssl_ca"] = ssl_ca
            if ssl_cert is not None:
                config["ssl_cert"] = ssl_cert
            if ssl_key is not None:
                config["ssl_key"] = ssl_key
        elif ssl is False:
            config["ssl_disabled"] = True

        if "collation" not in config:
            config["collation"] = "utf8mb4_general_ci"
        if "use_pure" not in config:
            config["use_pure"] = True
        try:
            connection = mysql.connector.connect(**config)
            connection.autocommit = True
            self.connection = connection
            return self.connection
        except mysql.connector.Error as e:
            logger.error(f"Error connecting to MySQL {self.database}, {e}!")
            raise

    def disconnect(self):
        """
        Closes the connection to the MySQL database if it's currently open.
        """
        if self.is_connected is False:
            return
        self.connection.close()
        return

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the MySQL database.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """

        result = StatusResponse(False)
        need_to_close = not self.is_connected

        try:
            connection = self.connect()
            result.success = connection.is_connected()
        except mysql.connector.Error as e:
            logger.error(f"Error connecting to MySQL {self.connection_data['database']}, {e}!")
            result.error_message = str(e)

        if result.success and need_to_close:
            self.disconnect()

        return result

    def native_query(self, query: str) -> Response:
        """
        Executes a SQL query on the MySQL database and returns the result.

        Args:
            query (str): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """

        need_to_close = not self.is_connected
        connection = None
        try:
            connection = self.connect()
            with connection.cursor(dictionary=True, buffered=True) as cur:
                cur.execute(query)
                if cur.with_rows:
                    result = cur.fetchall()
                    response = _make_table_response(result, cur)
                else:
                    response = Response(RESPONSE_TYPE.OK, affected_rows=cur.rowcount)
        except mysql.connector.Error as e:
            logger.error(f"Error running query: {query} on {self.connection_data['database']}!")
            response = Response(RESPONSE_TYPE.ERROR, error_message=str(e))
            if connection is not None and connection.is_connected():
                connection.rollback()

        if need_to_close:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement.
        """
        renderer = SqlalchemyRender("mysql")
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Get a list with all of the tabels in MySQL selected database
        """
        sql = """
            SELECT
                TABLE_SCHEMA AS table_schema,
                TABLE_NAME AS table_name,
                TABLE_TYPE AS table_type
            FROM
                information_schema.TABLES
            WHERE
                TABLE_TYPE IN ('BASE TABLE', 'VIEW')
                AND TABLE_SCHEMA = DATABASE()
            ORDER BY 2
            ;
        """
        result = self.native_query(sql)
        return result

    def get_columns(self, table_name) -> Response:
        """
        Show details about the table
        """
        q = f"""
            select
                COLUMN_NAME,
                DATA_TYPE,
                ORDINAL_POSITION,
                COLUMN_DEFAULT,
                IS_NULLABLE,
                CHARACTER_MAXIMUM_LENGTH,
                CHARACTER_OCTET_LENGTH,
                NUMERIC_PRECISION,
                NUMERIC_SCALE,
                DATETIME_PRECISION,
                CHARACTER_SET_NAME,
                COLLATION_NAME
            from
                information_schema.columns
            where
                table_name = '{table_name}';
        """
        result = self.native_query(q)
        result.to_columns_table_response(map_type_fn=_map_type)
        return result



"""del_log_table

Revision ID: 6ab9903fc59a
Revises: 4943359e354a
Create Date: 2025-02-10 16:50:27.186697

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa

# revision identifiers, used by Alembic.
revision = '6ab9903fc59a'
down_revision = '4943359e354a'
branch_labels = None
depends_on = None


def upgrade():
    op.drop_table('log')


def downgrade():
    op.create_table(
        'log',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.Column('created_at', sa.DATETIME(), nullable=True),
        sa.Column('log_type', sa.VARCHAR(), nullable=True),
        sa.Column('source', sa.VARCHAR(), nullable=True),
        sa.Column('company_id', sa.INTEGER(), nullable=True),
        sa.Column('payload', sa.VARCHAR(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )



import importlib

from mindsdb.interfaces.storage.model_fs import HandlerStorage


def create_validation_process(target: str, args: dict, integration_id: int, module_path: str) -> None:
    module = importlib.import_module(module_path)

    if hasattr(module.Handler, 'create_validation'):
        module.Handler.create_validation(
            target,
            args=args,
            handler_storage=HandlerStorage(integration_id)
        )



import os
import json
from typing import Dict, List, Literal, Tuple
from urllib.parse import urlparse

import pandas as pd
import psycopg
from mindsdb_sql_parser.ast import (
    Parameter,
    Identifier,
    BinaryOperation,
    Tuple as AstTuple,
    Constant,
    Select,
    OrderBy,
    TypeCast,
    Delete,
    Update,
    Function,
)
from pgvector.psycopg import register_vector

from mindsdb.integrations.handlers.postgres_handler.postgres_handler import (
    PostgresHandler,
)
from mindsdb.integrations.libs.response import RESPONSE_TYPE, HandlerResponse as Response
from mindsdb.integrations.libs.vectordatabase_handler import (
    FilterCondition,
    VectorStoreHandler,
    DistanceFunction,
    TableField,
    FilterOperator,
)
from mindsdb.integrations.libs.keyword_search_base import KeywordSearchBase
from mindsdb.integrations.utilities.sql_utils import KeywordSearchArgs
from mindsdb.utilities import log
from mindsdb.utilities.profiler import profiler
from mindsdb.utilities.context import context as ctx

logger = log.getLogger(__name__)


# todo Issue #7316 add support for different indexes and search algorithms e.g. cosine similarity or L2 norm
class PgVectorHandler(PostgresHandler, VectorStoreHandler, KeywordSearchBase):
    """This handler handles connection and execution of the PostgreSQL with pgvector extension statements."""

    name = "pgvector"

    def __init__(self, name: str, **kwargs):
        super().__init__(name=name, **kwargs)
        self._is_shared_db = False
        self._is_vector_registered = False
        # we get these from the connection args on PostgresHandler parent
        self._is_sparse = self.connection_args.get("is_sparse", False)
        self._vector_size = self.connection_args.get("vector_size", None)

        if self._is_sparse:
            if not self._vector_size:
                raise ValueError("vector_size is required when is_sparse=True")

                # Use inner product for sparse vectors
                distance_op = "<#>"

        else:
            distance_op = "<=>"
            if "distance" in self.connection_args:
                distance_ops = {
                    "l1": "<+>",
                    "l2": "<->",
                    "ip": "<#>",  # inner product
                    "cosine": "<=>",
                    "hamming": "<~>",
                    "jaccard": "<%>",
                }

                distance_op = distance_ops.get(self.connection_args["distance"])
                if distance_op is None:
                    raise ValueError(f"Wrong distance type. Allowed options are {list(distance_ops.keys())}")

        self.distance_op = distance_op
        self.connect()

    def get_metric_type(self) -> str:
        """
        Get the metric type from the distance ops

        """
        distance_ops_to_metric_type_map = {
            "<->": "vector_l2_ops",
            "<#>": "vector_ip_ops",
            "<=>": "vector_cosine_ops",
            "<+>": "vector_l1_ops",
            "<~>": "bit_hamming_ops",
            "<%>": "bit_jaccard_ops",
        }
        return distance_ops_to_metric_type_map.get(self.distance_op, "vector_cosine_ops")

    def _make_connection_args(self):
        cloud_pgvector_url = os.environ.get("KB_PGVECTOR_URL")
        # if no connection args and shared pg vector defined - use it
        if len(self.connection_args) == 0 and cloud_pgvector_url is not None:
            result = urlparse(cloud_pgvector_url)
            self.connection_args = {
                "host": result.hostname,
                "port": result.port,
                "user": result.username,
                "password": result.password,
                "database": result.path[1:],
            }
            self._is_shared_db = True
        return super()._make_connection_args()

    def get_tables(self) -> Response:
        # Hide list of tables from all users
        if self._is_shared_db:
            return Response(RESPONSE_TYPE.OK)
        return super().get_tables()

    def native_query(self, query, params=None) -> Response:
        # Prevent execute native queries
        if self._is_shared_db:
            return Response(RESPONSE_TYPE.OK)
        return super().native_query(query, params=params)

    def raw_query(self, query, params=None) -> Response:
        resp = super().native_query(query, params)
        if resp.resp_type == RESPONSE_TYPE.ERROR:
            raise RuntimeError(resp.error_message)
        if resp.resp_type == RESPONSE_TYPE.TABLE:
            return resp.data_frame

    @profiler.profile()
    def connect(self) -> psycopg.connection:
        """
        Handles the connection to a PostgreSQL database instance.
        """
        self.connection = super().connect()
        if self._is_vector_registered:
            return self.connection

        with self.connection.cursor() as cur:
            try:
                # load pg_vector extension
                cur.execute("CREATE EXTENSION IF NOT EXISTS vector")
                logger.info("pg_vector extension loaded")

            except psycopg.Error as e:
                self.connection.rollback()
                logger.error(f"Error loading pg_vector extension, ensure you have installed it before running, {e}!")
                raise

        # register vector type with psycopg2 connection
        register_vector(self.connection)
        self._is_vector_registered = True

        return self.connection

    def add_full_text_index(self, table_name: str, column_name: str) -> Response:
        """
        Add a full text index to the specified column of the table.
        Args:
            table_name (str): Name of the table to add the index to.
            column_name (str): Name of the column to add the index to.
        Returns:
            Response: Response object indicating success or failure.
        """
        table_name = self._check_table(table_name)
        query = f"CREATE INDEX IF NOT EXISTS {table_name}_{column_name}_fts_idx ON {table_name} USING gin(to_tsvector('english', {column_name}))"
        self.raw_query(query)
        return Response(RESPONSE_TYPE.OK)

    @staticmethod
    def _translate_conditions(conditions: List[FilterCondition]) -> Tuple[List[dict], dict]:
        """
        Translate filter conditions to a dictionary
        """

        if conditions is None:
            conditions = []

        filter_conditions = []
        embedding_condition = None

        for condition in conditions:
            is_embedding = condition.column == "embeddings"

            parts = condition.column.split(".")
            key = Identifier(parts[0])

            # converts 'col.el1.el2' to col->'el1'->>'el2'
            if len(parts) > 1:
                # intermediate elements
                for el in parts[1:-1]:
                    key = BinaryOperation(op="->", args=[key, Constant(el)])

                # last element
                key = BinaryOperation(op="->>", args=[key, Constant(parts[-1])])

            type_cast = None
            value = condition.value
            if (
                isinstance(value, list)
                and len(value) > 0
                and condition.op in (FilterOperator.IN, FilterOperator.NOT_IN)
            ):
                value = condition.value[0]

            if isinstance(value, int):
                type_cast = "int"
            elif isinstance(value, float):
                type_cast = "float"
            if type_cast is not None:
                key = TypeCast(type_cast, key)

            item = {
                "name": key,
                "op": condition.op.value,
                "value": condition.value,
            }
            if is_embedding:
                embedding_condition = item
            else:
                filter_conditions.append(item)

        return filter_conditions, embedding_condition

    @staticmethod
    def _construct_where_clause(filter_conditions=None):
        """
        Construct where clauses from filter conditions
        """

        where_clause = None

        for item in filter_conditions:
            key = item["name"]

            if item["op"].lower() in ("in", "not in"):
                values = [Constant(i) for i in item["value"]]
                value = AstTuple(values)
            else:
                value = Constant(item["value"])
            condition = BinaryOperation(op=item["op"], args=[key, value])

            if where_clause is None:
                where_clause = condition
            else:
                where_clause = BinaryOperation(op="AND", args=[where_clause, condition])
        return where_clause

    @staticmethod
    def _construct_full_after_from_clause(
        where_clause: str,
        offset_clause: str,
        limit_clause: str,
    ) -> str:
        return f"{where_clause} {offset_clause} {limit_clause}"

    def _build_keyword_bm25_query(
        self,
        table_name: str,
        keyword_search_args: KeywordSearchArgs,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        offset: int = None,
    ):
        if columns is None:
            columns = ["id", "content", "metadata"]

        filter_conditions, _ = self._translate_conditions(conditions)
        where_clause = self._construct_where_clause(filter_conditions)

        if keyword_search_args:
            keyword_query_condition = BinaryOperation(
                op="@@",
                args=[
                    Function("to_tsvector", args=[Constant("english"), Identifier(keyword_search_args.column)]),
                    Function("websearch_to_tsquery", args=[Constant("english"), Constant(keyword_search_args.query)]),
                ],
            )

            if where_clause:
                where_clause = BinaryOperation(op="AND", args=[where_clause, keyword_query_condition])
            else:
                where_clause = keyword_query_condition

        distance = Function(
            "ts_rank_cd",
            args=[
                Function("to_tsvector", args=[Constant("english"), Identifier(keyword_search_args.column)]),
                Function("websearch_to_tsquery", args=[Constant("english"), Constant(keyword_search_args.query)]),
            ],
            alias=Identifier("distance"),
        )

        targets = [Identifier(col) for col in columns]
        targets.append(distance)

        limit_clause = Constant(limit) if limit else None
        offset_clause = Constant(offset) if offset else None

        return Select(
            targets=targets,
            from_table=Identifier(table_name),
            where=where_clause,
            limit=limit_clause,
            offset=offset_clause,
        )

    def _build_select_query(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        offset: int = None,
    ) -> Select:
        """
        given inputs, build string query
        """
        limit_clause = Constant(limit) if limit else None
        offset_clause = Constant(offset) if offset else None

        # translate filter conditions to dictionary
        filter_conditions, embedding_search = self._translate_conditions(conditions)

        # given filter conditions, construct where clause
        where_clause = self._construct_where_clause(filter_conditions)

        # Handle distance column specially since it's calculated, not stored
        modified_columns = []
        has_distance = False
        if columns is not None:
            for col in columns:
                if col == TableField.DISTANCE.value:
                    has_distance = True
                else:
                    modified_columns.append(col)
        else:
            modified_columns = ["id", "content", "embeddings", "metadata"]
            has_distance = True

        targets = [Identifier(col) for col in modified_columns]

        query = Select(
            targets=targets,
            from_table=Identifier(table_name),
            where=where_clause,
            limit=limit_clause,
            offset=offset_clause,
        )

        if embedding_search:
            search_vector = embedding_search["value"]

            if self._is_sparse:
                # Convert dict to sparse vector if needed
                if isinstance(search_vector, dict):
                    from pgvector.utils import SparseVector

                    embedding = SparseVector(search_vector, self._vector_size)
                    search_vector = embedding.to_text()
            else:
                # Convert list to vector string if needed
                if isinstance(search_vector, list):
                    search_vector = f"[{','.join(str(x) for x in search_vector)}]"

            vector_op = BinaryOperation(
                op=self.distance_op,
                args=[Identifier("embeddings"), Constant(search_vector)],
                alias=Identifier("distance"),
            )
            # Calculate distance as part of the query if needed
            if has_distance:
                query.targets.append(vector_op)

            query.order_by = [OrderBy(vector_op, direction="ASC")]

        return query

    def _check_table(self, table_name: str):
        # Apply namespace for a user
        if self._is_shared_db:
            company_id = ctx.company_id or "x"
            return f"t_{company_id}_{table_name}"
        return table_name

    def select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
    ) -> pd.DataFrame:
        """
        Retrieve the data from the SQL statement with eliminated rows that dont satisfy the WHERE condition
        """
        table_name = self._check_table(table_name)

        if columns is None:
            columns = ["id", "content", "embeddings", "metadata"]

        query = self._build_select_query(table_name, columns, conditions, limit, offset)
        query_str = self.renderer.get_string(query, with_failback=True)
        result = self.raw_query(query_str)

        # ensure embeddings are returned as string so they can be parsed by mindsdb
        if "embeddings" in columns:
            result["embeddings"] = result["embeddings"].astype(str)

        return result

    def keyword_select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
        keyword_search_args: KeywordSearchArgs = None,
    ) -> pd.DataFrame:
        table_name = self._check_table(table_name)

        if columns is None:
            columns = ["id", "content", "embeddings", "metadata"]

        query = self._build_keyword_bm25_query(table_name, keyword_search_args, columns, conditions, limit, offset)
        query_str = self.renderer.get_string(query, with_failback=True)
        result = self.raw_query(query_str)

        # ensure embeddings are returned as string so they can be parsed by mindsdb
        if "embeddings" in columns:
            result["embeddings"] = result["embeddings"].astype(str)

        return result

    def hybrid_search(
        self,
        table_name: str,
        embeddings: List[float],
        query: str = None,
        metadata: Dict[str, str] = None,
        distance_function=DistanceFunction.COSINE_DISTANCE,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Executes a hybrid search, combining semantic search and one or both of keyword/metadata search.

        For insight on the query construction, see: https://docs.pgvecto.rs/use-case/hybrid-search.html#advanced-search-merge-the-results-of-full-text-search-and-vector-search.

        Args:
            table_name(str): Name of underlying table containing content, embeddings, & metadata
            embeddings(List[float]): Embedding vector to perform semantic search against
            query(str): User query to convert into keywords for keyword search
            metadata(Dict[str, str]): Metadata filters to filter content rows against
            distance_function(DistanceFunction): Distance function used to compare embeddings vectors for semantic search

        Kwargs:
            id_column_name(str): Name of ID column in underlying table
            content_column_name(str): Name of column containing document content in underlying table
            embeddings_column_name(str): Name of column containing embeddings vectors in underlying table
            metadata_column_name(str): Name of column containing metadata key-value pairs in underlying table

        Returns:
            df(pd.DataFrame): Hybrid search result, sorted by hybrid search rank
        """
        if query is None and metadata is None:
            raise ValueError(
                "Must provide at least one of: query for keyword search, or metadata filters. For only embeddings search, use normal search instead."
            )

        id_column_name = kwargs.get("id_column_name", "id")
        content_column_name = kwargs.get("content_column_name", "content")
        embeddings_column_name = kwargs.get("embeddings_column_name", "embeddings")
        metadata_column_name = kwargs.get("metadata_column_name", "metadata")
        # Filter by given metadata for semantic search & full text search CTEs, if present.
        where_clause = " WHERE "
        if metadata is None:
            where_clause = ""
            metadata = {}
        for i, (k, v) in enumerate(metadata.items()):
            where_clause += f"{metadata_column_name}->>'{k}' = '{v}'"
            if i < len(metadata.items()) - 1:
                where_clause += " AND "

        # See https://docs.pgvecto.rs/use-case/hybrid-search.html#advanced-search-merge-the-results-of-full-text-search-and-vector-search.
        #
        # We can break down the below query as follows:
        #
        # Start with a CTE (Common Table Expression) called semantic_search (https://www.postgresql.org/docs/current/queries-with.html).
        # This expression calculates rank by the defined distance function, which measures the distance between the
        # embeddings column and the given embeddings vector. Results are ordered by this rank.
        #
        # Next, define another CTE called full_text_search if we are doing keyword search.
        # This calculates rank using the built-in ts_rank function (https://www.postgresql.org/docs/current/textsearch-controls.html#TEXTSEARCH-RANKING).
        # We convert the content column to a ts_vector and match rows for the given tsquery in the content column. Results are ordered by this ts_rank.
        #
        # For both of these CTEs, we filter by any given metadata fields.
        #
        # See https://www.postgresql.org/docs/current/textsearch-controls.html#TEXTSEARCH-PARSING-DOCUMENTS for to_tsvector
        # See https://www.postgresql.org/docs/current/functions-textsearch.html#FUNCTIONS-TEXTSEARCH for tsquery syntax
        #
        # Finally, we use a FULL OUTER JOIN to SELECT from both CTEs defined above.
        # The COALESCE function is used to handle cases where one CTE has null values.
        #
        # Or, if we are only doing metadata search, we leave out the JOIN & full text search CTEs.
        #
        # We calculate the final "hybrid" rank by summing the reciprocals of the ranks from each individual CTE.
        semantic_search_cte = f"""WITH semantic_search AS (
    SELECT {id_column_name}, {content_column_name}, {embeddings_column_name},
    RANK () OVER (ORDER BY {embeddings_column_name} {distance_function.value} '{str(embeddings)}') AS rank
    FROM {table_name}{where_clause}
    ORDER BY {embeddings_column_name} {distance_function.value} '{str(embeddings)}'::vector
    )"""

        full_text_search_cte = ""
        if query is not None:
            ts_vector_clause = (
                f"WHERE to_tsvector('english', {content_column_name}) @@ plainto_tsquery('english', '{query}')"
            )
            if metadata:
                ts_vector_clause = (
                    f"AND to_tsvector('english', {content_column_name}) @@ plainto_tsquery('english', '{query}')"
                )
            full_text_search_cte = f""",
    full_text_search AS (
    SELECT {id_column_name}, {content_column_name}, {embeddings_column_name},
    RANK () OVER (ORDER BY ts_rank(to_tsvector('english', {content_column_name}), plainto_tsquery('english', '{query}')) DESC) AS rank
    FROM {table_name}{where_clause}
    {ts_vector_clause}
    ORDER BY ts_rank(to_tsvector('english', {content_column_name}), plainto_tsquery('english', '{query}')) DESC
    )"""

        hybrid_select = """
    SELECT * FROM semantic_search"""
        if query is not None:
            hybrid_select = f"""
    SELECT
        COALESCE(semantic_search.{id_column_name}, full_text_search.{id_column_name}) AS id,
        COALESCE(semantic_search.{content_column_name}, full_text_search.{content_column_name}) AS content,
        COALESCE(semantic_search.{embeddings_column_name}, full_text_search.{embeddings_column_name}) AS embeddings,
        COALESCE(1.0 / (1 + semantic_search.rank), 0.0) + COALESCE(1.0 / (1 + full_text_search.rank), 0.0) AS rank
    FROM semantic_search FULL OUTER JOIN full_text_search USING ({id_column_name}) ORDER BY rank DESC;
        """

        full_search_query = f"{semantic_search_cte}{full_text_search_cte}{hybrid_select}"
        return self.raw_query(full_search_query)

    def create_table(self, table_name: str):
        """Create a table with a vector column."""
        with self.connection.cursor() as cur:
            # For sparse vectors, use sparsevec type
            vector_column_type = "sparsevec" if self._is_sparse else "vector"

            # Vector size is required for sparse vectors, optional for dense
            if self._is_sparse and not self._vector_size:
                raise ValueError("vector_size is required for sparse vectors")

            # Add vector size specification only if provided
            size_spec = f"({self._vector_size})" if self._vector_size is not None else "()"
            if vector_column_type == "vector":
                size_spec = ""

            cur.execute(f"""
                CREATE TABLE IF NOT EXISTS {table_name} (
                    id TEXT PRIMARY KEY,
                    embeddings {vector_column_type}{size_spec},
                    content TEXT,
                    metadata JSONB
                )
            """)
            self.connection.commit()

    def insert(self, table_name: str, data: pd.DataFrame):
        """
        Insert data into the pgvector table database.
        """
        table_name = self._check_table(table_name)

        if "metadata" in data.columns:
            data["metadata"] = data["metadata"].apply(json.dumps)

        resp = super().insert(table_name, data)
        if resp.resp_type == RESPONSE_TYPE.ERROR:
            raise RuntimeError(resp.error_message)
        if resp.resp_type == RESPONSE_TYPE.TABLE:
            return resp.data_frame

    def update(self, table_name: str, data: pd.DataFrame, key_columns: List[str] = None):
        """
        Udate data into the pgvector table database.
        """
        table_name = self._check_table(table_name)

        where = None
        update_columns = {}

        for col in data.columns:
            value = Parameter("%s")

            if col in key_columns:
                cond = BinaryOperation(op="=", args=[Identifier(col), value])
                if where is None:
                    where = cond
                else:
                    where = BinaryOperation(op="AND", args=[where, cond])
            else:
                update_columns[col] = value

        query = Update(table=Identifier(table_name), update_columns=update_columns, where=where)

        if TableField.METADATA.value in data.columns:

            def fnc(v):
                if isinstance(v, dict):
                    return json.dumps(v)

            data[TableField.METADATA.value] = data[TableField.METADATA.value].apply(fnc)

            data = data.astype({TableField.METADATA.value: str})

        transposed_data = []
        for _, record in data.iterrows():
            row = [record[col] for col in update_columns.keys()]
            for key_column in key_columns:
                row.append(record[key_column])
            transposed_data.append(row)

        query_str = self.renderer.get_string(query)
        self.raw_query(query_str, transposed_data)

    def delete(self, table_name: str, conditions: List[FilterCondition] = None):
        table_name = self._check_table(table_name)

        filter_conditions, _ = self._translate_conditions(conditions)
        where_clause = self._construct_where_clause(filter_conditions)

        query = Delete(table=Identifier(table_name), where=where_clause)
        query_str = self.renderer.get_string(query, with_failback=True)
        self.raw_query(query_str)

    def drop_table(self, table_name: str, if_exists=True):
        """
        Run a drop table query on the pgvector database.
        """
        table_name = self._check_table(table_name)
        self.raw_query(f"DROP TABLE IF EXISTS {table_name}")

    def create_index(
        self,
        table_name: str,
        column_name: str = "embeddings",
        index_type: Literal["ivfflat", "hnsw"] = "hnsw",
        metric_type: str = None,
    ):
        """
        Create an index on the pgvector table.
        Args:
            table_name (str): Name of the table to create the index on.
            column_name (str): Name of the column to create the index on.
            index_type (str): Type of the index to create. Supported types are 'ivfflat' and 'hnsw'.
            metric_type (str): Metric type for the index. Supported types are 'vector_l2_ops', 'vector_ip_ops', and 'vector_cosine_ops'.
        """
        if metric_type is None:
            metric_type = self.get_metric_type()
        # Check if the index type is supported
        if index_type not in ["ivfflat", "hnsw"]:
            raise ValueError("Invalid index type. Supported types are 'ivfflat' and 'hnsw'.")
        table_name = self._check_table(table_name)
        # first we make sure embedding dimension is set
        embedding_dim_size_df = self.raw_query(f"SELECT vector_dims({column_name}) FROM {table_name} LIMIT 1")
        # check if answer is empty
        if embedding_dim_size_df.empty:
            raise ValueError("Could not determine embedding dimension size. Make sure that knowledge base isn't empty")
        try:
            embedding_dim = int(embedding_dim_size_df.iloc[0, 0])
            # alter table to add dimension
            self.raw_query(f"ALTER TABLE {table_name} ALTER COLUMN {column_name} TYPE vector({embedding_dim})")
        except Exception:
            raise ValueError("Could not determine embedding dimension size. Make sure that knowledge base isn't empty")

        # Create the index
        self.raw_query(f"CREATE INDEX ON {table_name} USING {index_type} ({column_name} {metric_type})")



"""update_project_names

Revision ID: 607709e1615b
Revises: 4c26ad04eeaa
Create Date: 2023-06-27 18:33:29.436607

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa


# revision identifiers, used by Alembic.
revision = '607709e1615b'
down_revision = 'b5bf593ba659'
branch_labels = None
depends_on = None


def upgrade():

    def _rename(table):
        conn = op.get_bind()

        data = conn.execute(
            table
            .select()
            .where(table.c.name.like("%.%"))
        ).fetchall()

        for row in data:
            name = row[0]
            name2 = name.replace('.', '_')

            op.execute(
                table
                .update()
                .where(table.c.name == name)
                .values({'name': name2})
            )

    projects = sa.Table(
        'project',
        sa.MetaData(),
        sa.Column('name', sa.String()),
    )

    _rename(projects)

    integrations = sa.Table(
        'integration',
        sa.MetaData(),
        sa.Column('name', sa.String()),
    )

    _rename(integrations)


def downgrade():
    pass



from typing import List, Dict, Any, Optional
import json
import pandas as pd
import ast

from mindsdb.interfaces.knowledge_base.preprocessing.models import (
    Document,
    ProcessedChunk,
    JSONChunkingConfig
)
from mindsdb.interfaces.knowledge_base.preprocessing.document_preprocessor import DocumentPreprocessor
from mindsdb.utilities import log

# Set up logger
logger = log.getLogger(__name__)


class JSONChunkingPreprocessor(DocumentPreprocessor):
    """JSON chunking preprocessor for handling JSON data structures"""

    def __init__(self, config: Optional[JSONChunkingConfig] = None):
        """Initialize with JSON chunking configuration"""
        super().__init__()
        self.config = config or JSONChunkingConfig()
        # No need for a text splitter here as we'll chunk by JSON structure

    def process_documents(self, documents: List[Document]) -> List[ProcessedChunk]:
        """Process JSON documents into chunks

        Args:
            documents: List of documents containing JSON content

        Returns:
            List of processed chunks
        """
        all_chunks = []

        for doc in documents:
            try:
                # Parse document content into a Python object
                json_data = self._parse_document_content(doc)
                if json_data is None:
                    # Handle parsing failure
                    error_message = "Content is neither valid JSON nor a valid Python literal."
                    error_chunk = self._create_error_chunk(doc, error_message)
                    all_chunks.append(error_chunk)
                    continue  # Skip to next document

                # Process the JSON data based on its structure
                chunks = self._process_json_data(json_data, doc)
                all_chunks.extend(chunks)
            except Exception as e:
                logger.error(f"Error processing document {doc.id}: {e}")
                error_chunk = self._create_error_chunk(doc, str(e))
                all_chunks.append(error_chunk)

        return all_chunks

    def _parse_document_content(self, doc: Document) -> Optional[Any]:
        """Parse document content into a Python object

        Args:
            doc: Document with content to parse

        Returns:
            Parsed content as a Python object or None if parsing failed
        """
        # If content is not a string, assume it's already a Python object
        if not isinstance(doc.content, str):
            return doc.content

        # Try to parse as JSON first
        try:
            return json.loads(doc.content)
        except json.JSONDecodeError:
            # If JSON parsing fails, try as Python literal
            try:
                return ast.literal_eval(doc.content)
            except (SyntaxError, ValueError) as e:
                logger.error(f"Error parsing content for document {doc.id}: {e}")
                # We'll create the error chunk in the main process_documents method
                return None

    def _process_json_data(self, json_data: Any, doc: Document) -> List[ProcessedChunk]:
        """Process JSON data based on its structure

        Args:
            json_data: Parsed JSON data as a Python object
            doc: Original document

        Returns:
            List of processed chunks
        """
        if isinstance(json_data, list):
            # List of objects - chunk by object
            return self._process_json_list(json_data, doc)
        elif isinstance(json_data, dict):
            # Single object - chunk according to config
            if self.config.chunk_by_object:
                return [self._create_chunk_from_dict(json_data, doc, 0, 1)]
            else:
                return self._process_json_dict(json_data, doc)
        else:
            # Primitive value - create a single chunk
            return [self._create_chunk_from_primitive(json_data, doc)]

    def _create_error_chunk(self, doc: Document, error_message: str) -> ProcessedChunk:
        """Create a chunk containing error information

        Args:
            doc: Original document
            error_message: Error message to include in the chunk

        Returns:
            ProcessedChunk with error information
        """
        return ProcessedChunk(
            id=f"{doc.id}_error",
            content=f"Error processing document: {error_message}",
            metadata=self._prepare_chunk_metadata(doc.id, 0, doc.metadata)
        )

    def _process_json_list(self, json_list: List, doc: Document) -> List[ProcessedChunk]:
        """Process a list of JSON objects into chunks"""
        chunks = []
        total_objects = len(json_list)

        for i, item in enumerate(json_list):
            if isinstance(item, dict):
                chunk = self._create_chunk_from_dict(item, doc, i, total_objects)
                chunks.append(chunk)
            elif isinstance(item, list):
                # Handle nested lists by converting to string representation
                chunk = self._create_chunk_from_primitive(
                    json.dumps(item),
                    doc,
                    chunk_index=i,
                    total_chunks=total_objects
                )
                chunks.append(chunk)
            else:
                # Handle primitive values
                chunk = self._create_chunk_from_primitive(
                    item,
                    doc,
                    chunk_index=i,
                    total_chunks=total_objects
                )
                chunks.append(chunk)

        return chunks

    def _process_json_dict(self, json_dict: Dict, doc: Document) -> List[ProcessedChunk]:
        """Process a single JSON object into chunks by fields"""
        chunks = []

        # Ensure we're working with a dictionary
        if isinstance(json_dict, str):
            try:
                json_dict = json.loads(json_dict)
            except json.JSONDecodeError:
                logger.error(f"Error parsing JSON string: {json_dict[:100]}...")
                return [self._create_error_chunk(doc, "Invalid JSON string")]

        # Filter fields based on include/exclude lists
        fields_to_process = {}
        for key, value in json_dict.items():
            if self.config.include_fields and key not in self.config.include_fields:
                continue
            if key in self.config.exclude_fields:
                continue
            fields_to_process[key] = value

        # Create a chunk for each field
        total_fields = len(fields_to_process)
        for i, (key, value) in enumerate(fields_to_process.items()):
            field_content = self._format_field_content(key, value)

            # Create chunk metadata
            metadata = self._prepare_chunk_metadata(doc.id, i, doc.metadata)
            metadata["field_name"] = key

            # Extract fields to metadata for filtering
            self._extract_fields_to_metadata(json_dict, metadata)

            # Generate chunk ID
            chunk_id = self._generate_chunk_id(
                chunk_index=i,
                total_chunks=total_fields,
                start_char=0,
                end_char=len(field_content),
                provided_id=doc.id,
                content_column=self.config.content_column
            )

            # Create and add the chunk
            chunk = ProcessedChunk(
                id=chunk_id,
                content=field_content,
                metadata=metadata
            )
            chunks.append(chunk)

        return chunks

    def _create_chunk_from_dict(self,
                                json_dict: Dict,
                                doc: Document,
                                chunk_index: int,
                                total_chunks: int) -> ProcessedChunk:
        """Create a chunk from a JSON dictionary"""
        # Ensure we're working with a dictionary
        if isinstance(json_dict, str):
            try:
                json_dict = json.loads(json_dict)
            except json.JSONDecodeError:
                logger.error(f"Error parsing JSON string: {json_dict[:100]}...")
                return self._create_error_chunk(doc, "Invalid JSON string")

        # Format the content
        if self.config.flatten_nested:
            flattened = self._flatten_dict(json_dict, self.config.nested_delimiter)
            filtered_dict = self._filter_fields(flattened)
            content = self._dict_to_text(filtered_dict)
        else:
            filtered_dict = {k: v for k, v in json_dict.items()
                             if (not self.config.include_fields or k in self.config.include_fields)
                             and k not in self.config.exclude_fields}
            content = json.dumps(filtered_dict, indent=2)

        # Create metadata
        metadata = self._prepare_chunk_metadata(doc.id, chunk_index, doc.metadata)

        # Extract fields to metadata for filtering
        self._extract_fields_to_metadata(json_dict, metadata)

        # Generate chunk ID
        chunk_id = self._generate_chunk_id(
            chunk_index=chunk_index,
            total_chunks=total_chunks,
            start_char=0,
            end_char=len(content),
            provided_id=doc.id,
            content_column=self.config.content_column
        )

        return ProcessedChunk(
            id=chunk_id,
            content=content,
            metadata=metadata
        )

    def _filter_fields(self, flattened_dict: Dict) -> Dict:
        """Filter fields based on include/exclude configuration"""
        # If include_fields is specified, only keep those fields
        if self.config.include_fields:
            filtered_dict = {k: v for k, v in flattened_dict.items()
                             if any(k == field or k.startswith(field + self.config.nested_delimiter)
                                    for field in self.config.include_fields)}
        else:
            filtered_dict = flattened_dict.copy()

        # Apply exclude_fields
        if self.config.exclude_fields:
            for exclude_field in self.config.exclude_fields:
                # Remove exact field match
                if exclude_field in filtered_dict:
                    filtered_dict.pop(exclude_field)

                # Remove any nested fields
                nested_prefix = exclude_field + self.config.nested_delimiter
                keys_to_remove = [k for k in filtered_dict if k.startswith(nested_prefix)]
                for key in keys_to_remove:
                    filtered_dict.pop(key)

        return filtered_dict

    def _create_chunk_from_primitive(
            self,
            value: Any,
            doc: Document,
            chunk_index: int = 0,
            total_chunks: int = 1
    ) -> ProcessedChunk:
        """Create a chunk from a primitive value"""
        content = str(value)

        # Create metadata
        metadata = self._prepare_chunk_metadata(doc.id, chunk_index, doc.metadata)

        # For primitive values, we don't have a JSON dictionary to extract fields from
        # But we can add the value itself as a metadata field if configured
        if self.config.extract_all_primitives:
            metadata["field_value"] = value

        # Generate chunk ID
        chunk_id = self._generate_chunk_id(
            chunk_index=chunk_index,
            total_chunks=total_chunks,
            start_char=0,
            end_char=len(content),
            provided_id=doc.id,
            content_column=self.config.content_column
        )

        return ProcessedChunk(
            id=chunk_id,
            content=content,
            metadata=metadata
        )

    def _flatten_dict(self, d: Dict, delimiter: str = '.', prefix: str = '') -> Dict:
        """Flatten a nested dictionary structure"""
        result = {}
        for k, v in d.items():
            new_key = f"{prefix}{delimiter}{k}" if prefix else k
            if isinstance(v, dict):
                result.update(self._flatten_dict(v, delimiter, new_key))
            elif isinstance(v, list) and all(isinstance(item, dict) for item in v):
                # Handle lists of dictionaries
                for i, item in enumerate(v):
                    result.update(self._flatten_dict(item, delimiter, f"{new_key}[{i}]"))
            else:
                result[new_key] = v
        return result

    def _dict_to_text(self, d: Dict) -> str:
        """Convert a dictionary to a human-readable text format"""
        lines = []
        for key, value in d.items():
            if value is None:
                continue
            if isinstance(value, list):
                if not value:
                    continue
                if all(isinstance(item, dict) for item in value):
                    # Format list of dictionaries
                    lines.append(f"{key}:")
                    for i, item in enumerate(value):
                        lines.append(f"  Item {i+1}:")
                        for k, v in item.items():
                            lines.append(f"    {k}: {v}")
                else:
                    # Format list of primitives
                    value_str = ", ".join(str(item) for item in value)
                    lines.append(f"{key}: {value_str}")
            else:
                lines.append(f"{key}: {value}")

        return "\n".join(lines)

    def _format_field_content(self, key: str, value: Any) -> str:
        """Format a field's content for inclusion in a chunk"""
        if isinstance(value, dict):
            if self.config.flatten_nested:
                flattened = self._flatten_dict(value, self.config.nested_delimiter, key)
                return self._dict_to_text(flattened)
            else:
                return f"{key}: {json.dumps(value, indent=2)}"
        elif isinstance(value, list):
            if all(isinstance(item, dict) for item in value):
                # Format list of dictionaries
                lines = [f"{key}:"]
                for i, item in enumerate(value):
                    lines.append(f"  Item {i+1}:")
                    for k, v in item.items():
                        lines.append(f"    {k}: {v}")
                return "\n".join(lines)
            else:
                # Format list of primitives
                value_str = ", ".join(str(item) for item in value if item is not None)
                return f"{key}: {value_str}"
        else:
            return f"{key}: {value}"

    def _extract_fields_to_metadata(self, json_dict: Dict, metadata: Dict) -> None:
        """Extract specified fields from JSON to metadata for filtering"""
        # Ensure we're working with a dictionary
        if isinstance(json_dict, str):
            try:
                json_dict = json.loads(json_dict)
            except json.JSONDecodeError:
                logger.error(f"Error parsing JSON string: {json_dict[:100]}...")
                return

        # Always flatten the dictionary for metadata extraction
        flattened = self._flatten_dict(json_dict, self.config.nested_delimiter)

        # If extract_all_primitives is True, extract all primitive values
        if self.config.extract_all_primitives:
            for key, value in flattened.items():
                if isinstance(value, (str, int, float, bool)) and value is not None:
                    metadata[f"field_{key}"] = value
            return

        # If metadata_fields is empty and extract_all_primitives is False,
        # assume all fields should be extracted
        if not self.config.metadata_fields:
            # First try to extract top-level primitive fields
            has_primitives = False
            for key, value in json_dict.items():
                if isinstance(value, (str, int, float, bool)) and value is not None:
                    metadata[f"field_{key}"] = value
                    has_primitives = True

            # If no top-level primitives were found, extract all primitives from flattened dict
            if not has_primitives:
                for key, value in flattened.items():
                    if isinstance(value, (str, int, float, bool)) and value is not None:
                        metadata[f"field_{key}"] = value
        else:
            # Extract only the specified fields
            for field in self.config.metadata_fields:
                if field in flattened and flattened[field] is not None:
                    metadata[f"field_{field}"] = flattened[field]
                else:
                    # Try to navigate the nested structure manually
                    parts = field.split(self.config.nested_delimiter)
                    current = json_dict
                    found = True

                    for part in parts:
                        if isinstance(current, dict) and part in current:
                            current = current[part]
                        else:
                            found = False
                            break

                    if found and current is not None:
                        metadata[f"field_{field}"] = current

    def to_dataframe(self, chunks: List[ProcessedChunk]) -> pd.DataFrame:
        """Convert processed chunks to dataframe format"""
        return pd.DataFrame([chunk.model_dump() for chunk in chunks])



import os
from google.auth.transport.requests import Request
from google.oauth2 import service_account
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import build
import pandas as pd
from pandas import DataFrame
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from .google_books_tables import BookshelvesTable, VolumesTable
from mindsdb.integrations.libs.api_handler import APIHandler, FuncParser
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class GoogleBooksHandler(APIHandler):
    """
    A class for handling connections and interactions with the Google Books API.
    """

    name = "google_books"

    def __init__(self, name: str, **kwargs):
        """
        Initialize the Google Books API handler.
        Args:
            name (str): name of the handler
            kwargs (dict): additional arguments
        """
        super().__init__(name)
        self.token = None
        self.service = None
        self.connection_data = kwargs.get("connection_data", {})
        self.credentials_file = self.connection_data.get("credentials", None)
        self.credentials = None
        self.scopes = ["https://www.googleapis.com/auth/books"]
        self.is_connected = False
        self.connection = None
        bookshelves = BookshelvesTable(self)
        self.bookshelves = bookshelves
        self._register_table("bookshelves", bookshelves)
        volumes = VolumesTable(self)
        self.volumes = volumes
        self._register_table("volumes", volumes)

    def connect(self):
        """
        Set up any connections required by the handler
        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.
        Returns:
            HandlerStatusResponse
        """

        if self.is_connected is True:
            return self.service
        if self.credentials_file:
            if os.path.exists("token_books.json"):
                self.credentials = Credentials.from_authorized_user_file("token_books.json", self.scopes)
            if not self.credentials or not self.credentials.valid:
                if self.credentials and self.credentials.expired and self.credentials.refresh_token:
                    self.credentials.refresh(Request())
                else:
                    self.credentials = service_account.Credentials.from_service_account_file(
                        self.credentials_file, scopes=self.scopes
                    )
            # Save the credentials for the next run
            with open("token_books.json", "w") as token:
                token.write(self.credentials.to_json())
            self.service = build("books", "v1", credentials=self.credentials)
        return self.service

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f"Error connecting to Google Books API: {e}!")
            response.error_message = e

        self.is_connected = response.success
        return response

    def native_query(self, query: str = None) -> Response:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                api's json etc)
        Returns:
            HandlerResponse
        """
        method_name, params = FuncParser().from_string(query)

        df = self.call_application_api(method_name, params)

        return Response(RESPONSE_TYPE.TABLE, data_frame=df)

    def get_bookshelves(self, params: dict = None) -> DataFrame:
        """
        Get bookshelf from Google Books API
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        minShelf = None
        maxShelf = None
        if params["shelf"]:
            shelf = int(params["shelf"])
            if params["source"]:
                df = (
                    service.mylibrary()
                    .bookshelves()
                    .get(shelf=shelf, userid=params["userid"], source=params["source"])
                    .execute()
                )
            else:
                df = service.mylibrary().bookshelves().get(shelf=shelf, userid=params["userid"]).execute()

            df = pd.DataFrame(df, columns=self.bookshelves.get_columns())
            return df
        elif not params["minShelf"] and params["maxShelf"]:
            minShelf = int(params["maxShelf"]) - 10
            maxShelf = int(params["maxShelf"])
        elif not params["maxShelf"] and params["minShelf"]:
            minShelf = int(params["minShelf"])
            maxShelf = int(params["minShelf"]) + 10
        elif params["maxShelf"] and params["minShelf"]:
            minShelf = int(params["minShelf"])
            maxShelf = int(params["maxShelf"])

        args = {
            key: value for key, value in params.items() if key not in ["minShelf", "maxShelf"] and value is not None
        }
        bookshelves = service.bookshelves().list(userid=params["userid"], **args).execute()

        df = pd.DataFrame(bookshelves["items"], columns=self.bookshelves.get_columns())
        if minShelf is not None or maxShelf is not None:
            # Drop bookshelves that are not in the id range
            df = df.drop(df[(df["id"] < minShelf) | (df["id"] > maxShelf)].index)

        return df

    def get_volumes(self, params: dict = None) -> DataFrame:
        """
        Get volumes from Google Books API
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        args = {key: value for key, value in params.items() if value is not None}
        volumes = service.volumes().list(**args).execute()
        df = pd.DataFrame(volumes["items"], columns=self.volumes.get_columns())
        return df

    def call_application_api(self, method_name: str = None, params: dict = None) -> DataFrame:
        """
        Call Google Books API and map the data to pandas DataFrame
        Args:
            method_name (str): method name
            params (dict): query parameters
        Returns:
            DataFrame
        """
        if method_name == "get_bookshelves":
            return self.get_bookshelves(params)
        elif method_name == "get_volumes":
            return self.get_volumes(params)
        else:
            raise NotImplementedError(f"Unknown method {method_name}")



import ast
from typing import Dict, Optional, List


from litellm import completion, batch_completion, embedding, acompletion, supports_response_schema

import pandas as pd

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log

from mindsdb.integrations.handlers.litellm_handler.settings import CompletionParameters


logger = log.getLogger(__name__)


class LiteLLMHandler(BaseMLEngine):
    """
    LiteLLMHandler is a MindsDB handler for litellm - https://docs.litellm.ai/docs/
    """

    name = "litellm"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if "using" not in args:
            raise Exception("Litellm engine requires a USING clause. See settings.py for more info on supported args.")

    @classmethod
    def prepare_arguments(cls, provider, model_name, args):
        if provider == "snowflake" and "snowflake_account_id" in args:
            args["api_base"] = (
                f"https://{args['snowflake_account_id']}.snowflakecomputing.com/api/v2/cortex/inference:complete"
            )
        if provider == "google":
            provider = "gemini"
        if "base_url" in args:
            args["api_base"] = args.pop("base_url")

        model_name = f"{provider}/{model_name}"
        return model_name, args

    @classmethod
    def embeddings(cls, provider: str, model: str, messages: List[str], args: dict) -> List[list]:
        model, args = cls.prepare_arguments(provider, model, args)
        response = embedding(model=model, input=messages, **args)
        return [rec["embedding"] for rec in response.data]

    @classmethod
    async def acompletion(cls, provider: str, model: str, messages: List[dict], args: dict):
        model, args = cls.prepare_arguments(provider, model, args)
        return await acompletion(model=model, messages=messages, stream=False, **args)

    @classmethod
    def completion(cls, provider: str, model: str, messages: List[dict], args: dict):
        model, args = cls.prepare_arguments(provider, model, args)
        json_output = args.pop("json_output", False)

        supports_json_output = supports_response_schema(model=model, custom_llm_provider=provider)

        if json_output and supports_json_output:
            args["response_format"] = {"type": "json_object"}
        else:
            args["response_format"] = None

        return completion(model=model, messages=messages, stream=False, **args)

    def create(
        self,
        target: str,
        df: pd.DataFrame = None,
        args: Optional[Dict] = None,
    ):
        """
        Dispatch is validating args and storing args in model_storage
        """
        # get api key from user input on create ML_ENGINE or create MODEL
        input_args = args["using"]

        # get api key from engine_storage
        ml_engine_args = self.engine_storage.get_connection_args()

        # check engine_storage for api_key
        input_args.update({k: v for k, v in ml_engine_args.items()})
        input_args["target"] = target

        # validate args
        export_args = CompletionParameters(**input_args).model_dump()

        # store args
        self.model_storage.json_set("args", export_args)

    def predict(self, df: pd.DataFrame = None, args: dict = None):
        """
        Dispatch is getting args from model_storage, validating args and running completion
        """

        input_args = self.model_storage.json_get("args")

        # validate args
        args = CompletionParameters(**input_args).model_dump()

        target = args.pop("target")

        # build messages
        self._build_messages(args, df)

        # remove prompt_template from args
        args.pop("prompt_template", None)

        if len(args["messages"]) > 1:
            # if more than one message, use batch completion
            responses = batch_completion(**args)
            return pd.DataFrame({target: [response.choices[0].message.content for response in responses]})

        # run completion
        response = completion(**args)

        return pd.DataFrame({target: [response.choices[0].message.content]})

    @staticmethod
    def _prompt_to_messages(prompt: str, **kwargs) -> List[Dict]:
        """
        Convert a prompt to a list of messages
        """

        if kwargs:
            # if kwargs are passed in, format the prompt with kwargs
            prompt = prompt.format(**kwargs)

        return [{"content": prompt, "role": "user"}]

    def _build_messages(self, args: dict, df: pd.DataFrame):
        """
        Build messages for completion
        """

        prompt_kwargs = df.iloc[0].to_dict()

        if "prompt_template" in prompt_kwargs:
            # if prompt_template is passed in predict query, use it
            logger.info(
                "Using 'prompt_template' passed in SELECT Predict query. "
                "Note this will overwrite a 'prompt_template' passed in create MODEL query."
            )

            args["prompt_template"] = prompt_kwargs.pop("prompt_template")

        if "mock_response" in prompt_kwargs:
            # used for testing to save on real completion api calls
            args["mock_response"]: str = prompt_kwargs.pop("mock_response")

        if "messages" in prompt_kwargs and len(prompt_kwargs) > 1:
            # if user passes in messages, no other args can be passed in
            raise Exception("If 'messages' is passed in SELECT Predict query, no other args can be passed in.")

        # if user passes in messages, use those instead
        if "messages" in prompt_kwargs:
            logger.info("Using messages passed in SELECT Predict query. 'prompt_template' will be ignored.")

            args["messages"]: List = ast.literal_eval(df["messages"].iloc[0])

        else:
            # if user passes in prompt_template, use that to create messages
            if len(prompt_kwargs) == 1:
                args["messages"] = (
                    self._prompt_to_messages(args["prompt_template"], **prompt_kwargs)
                    if args["prompt_template"]
                    else self._prompt_to_messages(df.iloc[0][0])
                )

            elif len(prompt_kwargs) > 1:
                try:
                    args["messages"] = self._prompt_to_messages(args["prompt_template"], **prompt_kwargs)
                except KeyError as e:
                    raise Exception(
                        f"{e}: Please pass in either a prompt_template on create MODEL or "
                        f"a single where clause in predict query."
                        f""
                    )



from typing import List, Optional
from pydantic import BaseModel, field_validator, model_validator
from pydantic_settings import BaseSettings


class LlamaIndexConfig(BaseSettings):
    """
    Model for LlamaIndexHandler settings.

    Attributes:
        default_index_class (str): Default index class.
        supported_index_class (List[str]): Supported index classes.
        default_reader (str): Default reader. Note this is custom data frame reader.
        supported_reader (List[str]): Supported readers.
    """
    DEFAULT_INDEX_CLASS: str = "VectorStoreIndex"
    SUPPORTED_INDEXES: List[str] = ["VectorStoreIndex"]
    DEFAULT_READER: str = "DFReader"
    SUPPORTED_READERS: List[str] = ["DFReader", "SimpleWebPageReader"]


llama_index_config = LlamaIndexConfig()


class LlamaIndexModel(BaseModel):
    """
    Model for LlamaIndexHandler.

    Attributes:
        reader (str): Reader.
        index_class (str): Index class.
        index (Any): Index.
        reader_params (Any): Reader parameters.
        index_params (Any): Index parameters.
    """
    reader: Optional[str] = None
    index_class: Optional[str] = None
    input_column: str
    openai_api_key: Optional[str] = None
    input_column: Optional[str]
    mode: Optional[str] = None
    user_column: Optional[str] = None
    assistant_column: Optional[str] = None

    @field_validator('reader')
    @classmethod
    def validate_reader(cls, value):
        if value not in llama_index_config.SUPPORTED_READERS:
            raise ValueError(f"Reader {value} is not supported.")

        return value

    @field_validator('index_class')
    @classmethod
    def validate_index_class(cls, value):
        if value not in llama_index_config.SUPPORTED_INDEXES:
            raise ValueError(f"Index class {value} is not supported.")

        return value

    @model_validator(mode='after')
    def validate_mode(self):
        if self.mode == "conversational" and not all([self.user_column, self.assistant_column]):
            raise ValueError("Conversational mode requires user_column and assistant_column parameter")

        return self



from typing import Text, List, Dict

from requests.exceptions import RequestException

from mindsdb.integrations.utilities.handlers.api_utilities.microsoft.ms_graph_api_utilities import MSGraphAPIBaseClient
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class MSGraphAPIOneDriveClient(MSGraphAPIBaseClient):
    """
    The Microsoft Graph API client for the Microsoft OneDrive handler.
    This client is used for accessing the Microsoft OneDrive specific endpoints of the Microsoft Graph API.
    Several common methods for submitting requests, fetching data, etc. are inherited from the base class.
    """

    def __init__(self, access_token: Text) -> None:
        super().__init__(access_token)

    def check_connection(self) -> bool:
        """
        Checks the connection to the Microsoft Graph API by fetching the user's profile.

        Returns:
            bool: True if the connection is successful, False otherwise.
        """
        try:
            self.fetch_paginated_data("me/drive")
            return True
        except RequestException as request_error:
            logger.error(f"Error checking connection: {request_error}")
            return False

    def get_all_items(self) -> List[Dict]:
        """
        Retrieves all items of the user's OneDrive.

        Returns:
            List[Dict]: All items of the user's OneDrive.
        """
        all_items = []
        for root_item in self.get_root_items():
            if "folder" in root_item:
                all_items.extend(self.get_child_items(root_item["id"], root_item["name"]))

            else:
                # Add the path to the item.
                root_item["path"] = root_item["name"]
                all_items.append(root_item)

        return all_items

    def get_root_items(self) -> List[Dict]:
        """
        Retrieves the root items of the user's OneDrive.

        Returns:
            List[Dict]: The root items of the user's OneDrive.
        """
        root_items = []
        for items in self.fetch_paginated_data("me/drive/root/children"):
            root_items.extend(items)

        return root_items

    def get_child_items(self, item_id: Text, path: Text) -> List[Dict]:
        """
        Recursively retrieves the child items of the specified item.

        Args:
            item_id (Text): The ID of the item whose child items are to be retrieved.

        Returns:
            List[Dict]: The child items of the specified item.
        """
        child_items = []
        for items in self.fetch_paginated_data(f"me/drive/items/{item_id}/children"):
            for item in items:
                child_path = f"{path}/{item['name']}"
                # If the item is a folder, get its child items.
                if "folder" in item:
                    # Recursively get the child items of the folder.
                    child_items.extend(self.get_child_items(item["id"], child_path))

                else:
                    # Add the path to the item.
                    item["path"] = child_path
                    child_items.append(item)

        return child_items

    def get_item_content(self, path: Text) -> bytes:
        """
        Retrieves the content of the specified item.

        Args:
            path (Text): The path of the item whose content is to be retrieved.

        Returns:
            bytes: The content of the specified item.
        """
        return self.fetch_data_content(f"me/drive/root:/{path}:/content")



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    repository={
        "type": ARG_TYPE.STR,
        "description": " GitHub repository name.",
        "required": True,
        "label": "Repository",
    },
    api_key={
        "type": ARG_TYPE.PWD,
        "description": "Optional GitHub API key to use for authentication.",
        "required": False,
        "label": "Api key",
        "secret": True
    },
    github_url={
        "type": ARG_TYPE.STR,
        "description": "Optional GitHub URL to connect to a GitHub Enterprise instance.",
        "required": False,
        "label": "Github url",
    },
)

connection_args_example = OrderedDict(
    repository="mindsdb/mindsdb",
    api_key="ghp_z91InCQZWZAMlddOzFCX7xHJrf9Fai35HT7",
    github_url="https://github.com/mindsdb/mindsdb"
)



from .__about__ import __description__ as description
from .__about__ import __version__ as version
from mindsdb.integrations.libs.const import HANDLER_TYPE
from .creation_args import creation_args
from .model_using_args import model_using_args

try:
    from .togetherai_handler import TogetherAIHandler as Handler
    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "TogetherAI"
name = "togetherai"
type = HANDLER_TYPE.ML
icon_path = "icon.svg"
permanent = False

__all__ = ["Handler", "version", "name", "type", "title", "description", "import_error", "icon_path", "creation_args", "model_using_args"]



from textwrap import dedent
from typing import Any
from contextlib import asynccontextmanager
from collections.abc import AsyncIterator
from dataclasses import dataclass

from mcp.server.fastmcp import FastMCP
from starlette.requests import Request
from starlette.responses import JSONResponse

from mindsdb.api.mysql.mysql_proxy.classes.fake_mysql_proxy import FakeMysqlProxy
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE as SQL_RESPONSE_TYPE
from mindsdb.interfaces.storage import db
from mindsdb.utilities import log

logger = log.getLogger(__name__)


@dataclass
class AppContext:
    db: Any


@asynccontextmanager
async def app_lifespan(server: FastMCP) -> AsyncIterator[AppContext]:
    """Manage application lifecycle with type-safe context"""
    # Initialize on startup
    db.init()
    try:
        yield AppContext(db=db)
    finally:
        # TODO: We need better way to handle this in storage/db.py
        pass


# Configure server with lifespan
mcp = FastMCP(
    "MindsDB",
    lifespan=app_lifespan,
    dependencies=["mindsdb"],  # Add any additional dependencies
)


# MCP Queries
LISTING_QUERY = "SHOW DATABASES"


query_tool_description = dedent("""\
    Executes a SQL query against MindsDB.

    A database must be specified either in the `context` parameter or directly in the query string (e.g., `SELECT * FROM my_database.my_table`). Queries like `SELECT * FROM my_table` will fail without a `context`.

    Args:
        query (str): The SQL query to execute.
        context (dict, optional): The default database context. For example, `{"db": "my_postgres"}`.

    Returns:
        A dictionary describing the result.
        - For a successful query with no data to return (e.g., an `UPDATE` statement), the response is `{"type": "ok"}`.
        - If the query returns tabular data, the response is a dictionary containing `data` (a list of rows) and `column_names` (a list of column names). For example: `{"type": "table", "data": [[1, "a"], [2, "b"]], "column_names": ["column_a", "column_b"]}`.
        - In case of an error, a response is `{"type": "error", "error_message": "the error message"}`.
""")


@mcp.tool(name="query", description=query_tool_description)
def query(query: str, context: dict | None = None) -> dict[str, Any]:
    """Execute a SQL query against MindsDB

    Args:
        query: The SQL query to execute
        context: Optional context parameters for the query

    Returns:
        Dict containing the query results or error information
    """

    if context is None:
        context = {}

    logger.debug(f"Incoming MCP query: {query}")

    mysql_proxy = FakeMysqlProxy()
    mysql_proxy.set_context(context)

    try:
        result = mysql_proxy.process_query(query)

        if result.type == SQL_RESPONSE_TYPE.OK:
            return {"type": SQL_RESPONSE_TYPE.OK}

        if result.type == SQL_RESPONSE_TYPE.TABLE:
            return {
                "type": SQL_RESPONSE_TYPE.TABLE,
                "data": result.result_set.to_lists(json_types=True),
                "column_names": [column.alias or column.name for column in result.result_set.columns],
            }
        else:
            return {"type": SQL_RESPONSE_TYPE.ERROR, "error_code": 0, "error_message": "Unknown response type"}

    except Exception as e:
        logger.error(f"Error processing query: {str(e)}")
        return {"type": SQL_RESPONSE_TYPE.ERROR, "error_code": 0, "error_message": str(e)}


list_databases_tool_description = (
    "Returns a list of all database connections currently available in MindsDB. "
    + "The tool takes no parameters and responds with a list of database names, "
    + 'for example: ["my_postgres", "my_mysql", "test_db"].'
)


@mcp.tool(name="list_databases", description=list_databases_tool_description)
def list_databases() -> list[str]:
    """
    List all databases in MindsDB

    Returns:
       list[str]: list of databases
    """

    mysql_proxy = FakeMysqlProxy()

    try:
        result = mysql_proxy.process_query(LISTING_QUERY)
        if result.type == SQL_RESPONSE_TYPE.ERROR:
            return {
                "type": "error",
                "error_code": result.error_code,
                "error_message": result.error_message,
            }

        elif result.type == SQL_RESPONSE_TYPE.OK:
            return {"type": "ok"}

        elif result.type == SQL_RESPONSE_TYPE.TABLE:
            data = result.result_set.to_lists(json_types=True)
            data = [val[0] for val in data]
            return data

    except Exception as e:
        return {
            "type": "error",
            "error_code": 0,
            "error_message": str(e),
        }


def _get_status(request: Request) -> JSONResponse:
    """
    Status endpoint that returns basic server information.
    This endpoint can be used by the frontend to check if the MCP server is running.
    """

    status_info = {
        "status": "ok",
        "service": "mindsdb-mcp",
    }

    return JSONResponse(status_info)


def get_mcp_app():
    app = mcp.sse_app()
    app.add_route("/status", _get_status, methods=["GET"])
    return app



import os
from typing import Callable

from mindsdb_sql_parser import Function, Constant, Variable

from mindsdb.utilities import log
from mindsdb.interfaces.storage.fs import RESOURCE_GROUP
from mindsdb.interfaces.storage.json import get_json_storage


logger = log.getLogger(__name__)


ENV_VAR_PREFIX = "MDB_"


class VariablesController:
    def __init__(self) -> None:
        self._storage = get_json_storage(resource_id=0, resource_group=RESOURCE_GROUP.SYSTEM)
        self._store_key = "variables"
        self._data = None

    def _get_data(self) -> dict:
        if self._data is None:
            self._data = self._storage.get(self._store_key)
            if self._data is None:
                self._data = {}
        return self._data

    def get_value(self, name: str):
        data = self._get_data()
        if name not in data:
            raise ValueError(f"Variable {name} is not defined")
        return data[name]

    def set_value(self, name: str, value):
        data = self._get_data()
        data[name] = value
        self._storage.set(self._store_key, data)

    def _from_env(self, name: Constant) -> str:
        # gets variable value from environment.
        # available names are restricted by ENV_VAR_PREFIX to don't provide access to arbitrary venv variable

        var_name = name.value
        if not var_name.startswith(ENV_VAR_PREFIX):
            raise ValueError(f"Can access only to variable names starting with {ENV_VAR_PREFIX}")
        if var_name not in os.environ:
            raise ValueError(f"Environment variable {var_name} is not defined")
        return os.environ[var_name]

    def _get_function(self, name: str) -> Callable:
        if name == "from_env":
            return self._from_env
        raise ValueError(f"Function {name} is not found")

    def set_variable(self, name: str, value):
        # store new value for variable in database
        #   if value is a function - extract value using this function

        name = name.lower()
        if isinstance(value, Function):
            fnc = self._get_function(value.op)
            value = fnc(*value.args)

        elif isinstance(value, Constant):
            value = value.value

        else:
            # ignore
            return

        self.set_value(name, value)

    def fill_parameters(self, var):
        # recursively check input and fill Variables if they exist there

        if isinstance(var, Variable):
            return self.get_value(var.value.lower())
        if isinstance(var, Function):
            fnc = self._get_function(var.op)
            return fnc(*var.args)
        elif isinstance(var, dict):
            return {key: self.fill_parameters(value) for key, value in var.items()}
        elif isinstance(var, list):
            return [self.fill_parameters(value) for value in var]
        return var


variables_controller = VariablesController()



from typing import Text, List, Dict
import random
import time
import math

import openai

import tiktoken

import mindsdb.utilities.profiler as profiler


class PendingFT(openai.OpenAIError):
    """
    Custom exception to handle pending fine-tuning status.
    """

    message: str

    def __init__(self, message) -> None:
        super().__init__()
        self.message = message


def retry_with_exponential_backoff(
    initial_delay: float = 1,
    hour_budget: float = 0.3,
    jitter: bool = False,
    exponential_base: int = 2,
    wait_errors: tuple = (openai.APITimeoutError, openai.APIConnectionError, PendingFT),
    status_errors: tuple = (openai.APIStatusError, openai.APIResponseValidationError),
):
    """
    Wrapper to enable optional arguments. It means this decorator always needs to be called with parenthesis:

    > @retry_with_exponential_backoff()  # optional argument override here
    > def f(): [...]

    """  # noqa

    @profiler.profile()
    def _retry_with_exponential_backoff(func):
        """
        Exponential backoff to retry requests on a rate-limited API call, as recommended by OpenAI.
        Loops the call until a successful response or max_retries is hit or an exception is raised.

        Slight changes in the implementation, but originally from:
        https://github.com/openai/openai-cookbook/blob/main/examples/How_to_handle_rate_limits.ipynb

        Args:
            func: Function to be wrapped
            initial_delay: Initial delay in seconds
            hour_budget: Hourly budget in seconds
            jitter: Adds randomness to the delay
            exponential_base: Base for the exponential backoff
            wait_errors: Tuple of errors to retry on
            status_errors: Tuple of status errors to raise

        Returns:
            Wrapper function with exponential backoff
        """  # noqa

        def wrapper(*args, **kwargs):
            num_retries = 0
            delay = initial_delay

            if isinstance(hour_budget, float) or isinstance(hour_budget, int):
                try:
                    max_retries = round((math.log((hour_budget * 3600) / initial_delay)) / math.log(exponential_base))
                except ValueError:
                    max_retries = 10
            else:
                max_retries = 10
            max_retries = max(1, max_retries)

            while True:
                try:
                    return func(*args, **kwargs)

                except status_errors as e:
                    raise Exception(
                        f"Error status {e.status_code} raised by OpenAI API: {e.body.get('message', 'Please refer to `https://platform.openai.com/docs/guides/error-codes` for more information.')}"  # noqa
                    )  # noqa

                except wait_errors:
                    num_retries += 1
                    if num_retries > max_retries:
                        raise Exception(f"Maximum number of retries ({max_retries}) exceeded.")
                    # Increment the delay and wait
                    delay *= exponential_base * (1 + jitter * random.random())
                    time.sleep(delay)

                except openai.OpenAIError as e:
                    raise Exception(
                        f"General {str(e)} error raised by OpenAI. Please refer to `https://platform.openai.com/docs/guides/error-codes` for more information."  # noqa
                    )

                except Exception as e:
                    raise e

        return wrapper

    return _retry_with_exponential_backoff


def truncate_msgs_for_token_limit(messages: List[Dict], model_name: Text, max_tokens: int, truncate: Text = "first"):
    """
    Truncates message list to fit within the token limit.
    The first message for chat completion models are general directives with the system role, which will ideally be kept at all times.

    Slight changes in the implementation, but originally from:
    https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb

    Args:
        messages (List[Dict]): List of messages
        model_name (Text): Model name
        max_tokens (int): Maximum token limit
        truncate (Text): Truncate strategy, either 'first' or 'last'

    Returns:
        List[Dict]: Truncated message list
    """  # noqa
    try:
        encoder = tiktoken.encoding_for_model(model_name)
    except KeyError:
        # If the encoding is not found, defualt to cl100k_base.
        # This is applicable for handlers that extend the OpenAI handler such as Anyscale.
        model_name = "gpt-3.5-turbo-0301"
        encoder = tiktoken.get_encoding("cl100k_base")

    sys_priming = messages[0:1]
    n_tokens = count_tokens(messages, encoder, model_name)
    while n_tokens > max_tokens:
        if len(messages) == 2:
            return messages[:-1]  # edge case: if limit is surpassed by just one input, we remove initial instruction
        elif len(messages) == 1:
            return messages

        if truncate == "first":
            messages = sys_priming + messages[2:]
        else:
            messages = sys_priming + messages[1:-1]

        n_tokens = count_tokens(messages, encoder, model_name)
    return messages


def count_tokens(messages: List[Dict], encoder: tiktoken.core.Encoding, model_name: Text = "gpt-3.5-turbo-0301"):
    """
    Counts the number of tokens in a list of messages.

    Args:
        messages: List of messages
        encoder: Tokenizer
        model_name: Model name
    """
    if "gpt-3.5-turbo" in model_name:  # note: future models may deviate from this (only 0301 really complies)
        tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        tokens_per_name = -1
    else:
        tokens_per_message = 3
        tokens_per_name = 1

    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message

        for key, value in message.items():
            num_tokens += len(encoder.encode(value))
            if key == "name":  # if there's a name, the role is omitted
                num_tokens += tokens_per_name
    num_tokens += 2  # every reply is primed with <im_start>assistant
    return num_tokens


def get_available_models(client) -> List[Text]:
    """
    Returns a list of available openai models for the given API key.

    Args:
        client: openai sdk client

    Returns:
        List[Text]: List of available models
    """
    res = client.models.list()

    return [models.id for models in res.data]



from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import JSONResponse
from starlette.requests import Request
from http import HTTPStatus
from typing import Optional
import secrets
import hmac
import hashlib
import os
import traceback

from mindsdb.utilities import log
from mindsdb.utilities.config import config

logger = log.getLogger(__name__)

SECRET_KEY = os.environ.get("AUTH_SECRET_KEY") or secrets.token_urlsafe(32)
# We store token (fingerprints) in memory, which means everyone is logged out if the process restarts
TOKENS = []


def get_pat_fingerprint(token: str) -> str:
    """Hash the token with HMAC-SHA256 using secret_key as pepper."""
    return hmac.new(SECRET_KEY.encode(), token.encode(), hashlib.sha256).hexdigest()


def generate_pat() -> str:
    logger.debug("Generating new auth token")
    token = "pat_" + secrets.token_urlsafe(32)
    TOKENS.append(get_pat_fingerprint(token))
    return token


def verify_pat(raw_token: str) -> bool:
    """Verify if the raw_token matches a stored fingerprint.
    Returns token_id if valid, None if not.
    """
    if not raw_token:
        return False
    fp = get_pat_fingerprint(raw_token)
    for stored_fp in TOKENS:
        if hmac.compare_digest(fp, stored_fp):
            return True
    return False


def revoke_pat(raw_token: str) -> bool:
    """Revoke raw_token from active tokens"""
    if not raw_token:
        return False
    fp = get_pat_fingerprint(raw_token)
    for stored_fp in TOKENS:
        if hmac.compare_digest(fp, stored_fp):
            TOKENS.remove(stored_fp)
            return True
    return False


class PATAuthMiddleware(BaseHTTPMiddleware):
    def _extract_bearer(self, request: Request) -> Optional[str]:
        h = request.headers.get("Authorization")
        if not h or not h.startswith("Bearer "):
            return None
        return h.split(" ", 1)[1].strip() or None

    async def dispatch(self, request: Request, call_next):
        if config.get("auth", {}).get("http_auth_enabled", False) is False:
            return await call_next(request)

        token = self._extract_bearer(request)
        if not token or not verify_pat(token):
            return JSONResponse({"detail": "Unauthorized"}, status_code=HTTPStatus.UNAUTHORIZED)

        request.state.user = config["auth"].get("username")
        return await call_next(request)


# Used by mysql and postgres protocols
def check_auth(username, password, scramble_func, salt, company_id, config):
    try:
        hardcoded_user = config["auth"].get("username")
        hardcoded_password = config["auth"].get("password")
        if hardcoded_password is None:
            hardcoded_password = ""
        hardcoded_password_hash = scramble_func(hardcoded_password, salt)
        hardcoded_password = hardcoded_password.encode()

        if password is None:
            password = ""
        if isinstance(password, str):
            password = password.encode()

        if username != hardcoded_user:
            logger.warning(f"Check auth, user={username}: user mismatch")
            return {"success": False}

        if password != hardcoded_password and password != hardcoded_password_hash:
            logger.warning(f"check auth, user={username}: password mismatch")
            return {"success": False}

        logger.info(f"Check auth, user={username}: Ok")
        return {"success": True, "username": username}
    except Exception as e:
        logger.error(f"Check auth, user={username}: ERROR")
        logger.error(e)
        logger.error(traceback.format_exc())



from typing import Text, Dict, Optional, Any

import ibm_db_dbi
from ibm_db_dbi import OperationalError, ProgrammingError
from ibm_db_sa.ibm_db import DB2Dialect_ibm_db as DB2Dialect
from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
import pandas as pd

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class DB2Handler(DatabaseHandler):
    name = "db2"

    def __init__(self, name: Text, connection_data: Optional[Dict], **kwargs: Any) -> None:
        """
        Initializes the handler.
        Args:
            name (Text): The name of the handler instance.
            connection_data (Dict): The connection data required to connect to the IBM Db2 database.
            kwargs: Arbitrary keyword arguments.
        """
        super().__init__(name)
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self) -> None:
        """
        Closes the connection when the handler instance is deleted.
        """
        if self.is_connected:
            self.disconnect()

    def connect(self) -> ibm_db_dbi.Connection:
        """
        Establishes a connection to a IBM Db2 database.

        Raises:
            ValueError: If the required connection parameters are not provided.
            ibm_db_dbi.OperationalError: If an error occurs while connecting to the IBM Db2 database.

        Returns:
            ibm_db_dbi.Connection: A connection object to the IBM Db2 database.
        """
        if self.is_connected:
            return self.connection

        # Mandatory connection parameters.
        if not all(key in self.connection_data for key in ["host", "user", "password", "database"]):
            raise ValueError("Required parameters (host, user, password, database) must be provided.")
        cloud = "databases.appdomain.cloud" in self.connection_data["host"]
        if cloud:
            connection_string = f"DATABASE={self.connection_data['database']};HOSTNAME={self.connection_data['host']};PORT={self.connection_data['port']};PROTOCOL=TCPIP;UID={self.connection_data['user']};PWD={self.connection_data['password']};SECURITY=SSL;"
            connection_string += "SSLSERVERCERTIFICATE=;"
        else:
            connection_string = f"DRIVER={'IBM DB2 ODBC DRIVER'};DATABASE={self.connection_data['database']};HOST={self.connection_data['host']};PROTOCOL=TCPIP;UID={self.connection_data['user']};PWD={self.connection_data['password']};"

            # Optional connection parameters.
            if "port" in self.connection_data:
                connection_string += f"PORT={self.connection_data['port']};"

        if "schema" in self.connection_data:
            connection_string += f"CURRENTSCHEMA={self.connection_data['schema']};"

        try:
            self.connection = ibm_db_dbi.pconnect(connection_string, "", "")
            self.is_connected = True
            return self.connection
        except OperationalError as operational_error:
            logger.error(f"Error while connecting to {self.connection_data.get('database')}, {operational_error}!")
            raise
        except Exception as unknown_error:
            logger.error(f"Unknown error while connecting to {self.connection_data.get('database')}, {unknown_error}!")
            raise

    def disconnect(self) -> None:
        """
        Closes the connection to the IBM Db2 database if it's currently open.
        """
        if not self.is_connected:
            return

        self.connection.close()
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the IBM Db2 database.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except (OperationalError, ValueError) as known_error:
            logger.error(f"Connection check to IBM Db2 failed, {known_error}!")
            response.error_message = str(known_error)
        except Exception as unknown_error:
            logger.error(f"Connection check to IBM Db2 failed due to an unknown error, {unknown_error}!")
            response.error_message = str(unknown_error)

        if response.success and need_to_close:
            self.disconnect()

        elif not response.success and self.is_connected:
            self.is_connected = False

        return response

    def native_query(self, query: Text) -> Response:
        """
        Executes a SQL query on the IBM Db2 database and returns the result (if any).

        Args:
            query (str): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """
        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor() as cur:
            try:
                cur.execute(query)

                if cur._result_set_produced:
                    result = cur.fetchall()
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame(result, columns=[x[0] for x in cur.description]),
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)
                connection.commit()
            except (OperationalError, ProgrammingError) as known_error:
                logger.error(f"Error running query: {query} on {self.connection_data.get('database')}!")
                response = Response(RESPONSE_TYPE.ERROR, error_message=str(known_error))
                connection.rollback()

            except Exception as unknown_error:
                logger.error(f"Unknown error running query: {query} on {self.connection_data.get('database')}!")
                response = Response(RESPONSE_TYPE.ERROR, error_message=str(unknown_error))
                connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode on the IBM Db2 database and retrieves the data (if any).

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Returns:
            Response: The response from the `native_query` method, containing the result of the SQL query execution.
        """
        renderer = SqlalchemyRender(DB2Dialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Retrieves a list of all non-system tables and views in the current schema of the IBM Db2 database.

        Returns:
            Response: A response object containing the list of tables and views, formatted as per the `Response` class.
        """
        connection = self.connect()

        result = connection.tables(connection.current_schema)

        tables = []
        for table in result:
            tables.append(
                {
                    "TABLE_NAME": table["TABLE_NAME"],
                    "TABLE_SCHEMA": table["TABLE_SCHEM"],
                    "TABLE_TYPE": table["TABLE_TYPE"],
                }
            )

        response = Response(RESPONSE_TYPE.TABLE, data_frame=pd.DataFrame(tables))

        return response

    def get_columns(self, table_name: Text) -> Response:
        """
        Retrieves column details for a specified table in the IBM Db2 database.

        Args:
            table_name (Text): The name of the table for which to retrieve column information.

        Raises:
            ValueError: If the 'table_name' is not a valid string.

        Returns:
            Response: A response object containing the column details.
        """
        if not table_name or not isinstance(table_name, str):
            raise ValueError("Invalid table name provided.")

        connection = self.connect()

        result = connection.columns(table_name=table_name)

        columns = [column["COLUMN_NAME"] for column in result]

        response = Response(RESPONSE_TYPE.TABLE, data_frame=pd.DataFrame(columns, columns=["COLUMN_NAME"]))

        return response



import asyncio
import argparse
import json

from typing import List, Dict, Optional
from contextlib import AsyncExitStack

import uvicorn
from fastapi import FastAPI, HTTPException, BackgroundTasks
from fastapi.responses import StreamingResponse
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel, Field
from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client

from mindsdb.utilities import log
from mindsdb.interfaces.agents.mcp_client_agent import create_mcp_agent

logger = log.getLogger(__name__)

app = FastAPI(title="MindsDB MCP Agent LiteLLM API")

# Configure CORS
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# Store agent wrapper as a global variable
agent_wrapper = None
# MCP session for direct SQL queries
mcp_session = None
exit_stack = AsyncExitStack()


class ChatMessage(BaseModel):
    role: str
    content: str


class ChatCompletionRequest(BaseModel):
    model: str
    messages: List[ChatMessage]
    stream: bool = False
    temperature: Optional[float] = None
    max_tokens: Optional[int] = None


class ChatCompletionChoice(BaseModel):
    index: int = 0
    message: Optional[Dict[str, str]] = None
    delta: Optional[Dict[str, str]] = None
    finish_reason: Optional[str] = "stop"


class ChatCompletionResponse(BaseModel):
    id: str = "mcp-agent-response"
    object: str = "chat.completion"
    created: int = 0
    model: str
    choices: List[ChatCompletionChoice]
    usage: Dict[str, int] = Field(default_factory=lambda: {"prompt_tokens": 0, "completion_tokens": 0, "total_tokens": 0})


class DirectSQLRequest(BaseModel):
    query: str


@app.post("/v1/chat/completions")
async def chat_completions(request: ChatCompletionRequest):
    global agent_wrapper

    if agent_wrapper is None:
        raise HTTPException(status_code=500, detail="Agent not initialized. Make sure MindsDB server is running with MCP enabled: python -m mindsdb --api=mysql,mcp,http")

    try:
        # Convert request to messages format
        messages = [
            {"role": msg.role, "content": msg.content}
            for msg in request.messages
        ]

        if request.stream:
            # Return a streaming response
            async def generate():
                try:
                    async for chunk in agent_wrapper.acompletion_stream(messages, model=request.model):
                        yield f"data: {json.dumps(chunk)}\n\n"
                    yield "data: [DONE]\n\n"
                except Exception as e:
                    logger.error(f"Streaming error: {str(e)}")
                    yield "data: {{'error': 'Streaming failed due to an internal error.'}}\n\n"
            return StreamingResponse(generate(), media_type="text/event-stream")
        else:
            # Return a regular response
            response = await agent_wrapper.acompletion(messages)

            # Ensure the content is a string
            content = response["choices"][0]["message"].get("content", "")
            if not isinstance(content, str):
                content = str(content)

            # Transform to proper OpenAI format
            return ChatCompletionResponse(
                model=request.model,
                choices=[
                    ChatCompletionChoice(
                        message={"role": "assistant", "content": content}
                    )
                ]
            )

    except Exception as e:
        logger.error(f"Error in chat completion: {str(e)}")
        raise HTTPException(status_code=500, detail=str(e))


@app.post("/direct-sql")
async def direct_sql(request: DirectSQLRequest, background_tasks: BackgroundTasks):
    """Execute a direct SQL query via MCP (for testing)"""
    global agent_wrapper, mcp_session

    if agent_wrapper is None and mcp_session is None:
        raise HTTPException(status_code=500, detail="No MCP session available. Make sure MindsDB server is running with MCP enabled.")

    try:
        # First try to use the agent's session if available
        if hasattr(agent_wrapper.agent, "session") and agent_wrapper.agent.session:
            session = agent_wrapper.agent.session
            result = await session.call_tool("query", {"query": request.query})
            return {"result": result.content}
        # If agent session not available, use the direct session
        elif mcp_session:
            result = await mcp_session.call_tool("query", {"query": request.query})
            return {"result": result.content}
        else:
            raise HTTPException(status_code=500, detail="No MCP session available")

    except Exception as e:
        logger.error(f"Error executing direct SQL: {str(e)}")
        raise HTTPException(status_code=500, detail=str(e))


@app.get("/v1/models")
async def list_models():
    """List available models - always returns the single model we're using"""
    global agent_wrapper

    if agent_wrapper is None:
        return {
            "object": "list",
            "data": [
                {
                    "id": "mcp-agent",
                    "object": "model",
                    "created": 0,
                    "owned_by": "mindsdb"
                }
            ]
        }

    # Return the actual model name if available
    model_name = agent_wrapper.agent.args.get("model_name", "mcp-agent")

    return {
        "object": "list",
        "data": [
            {
                "id": model_name,
                "object": "model",
                "created": 0,
                "owned_by": "mindsdb"
            }
        ]
    }


@app.get("/health")
async def health_check():
    """Health check endpoint"""
    global agent_wrapper

    health_status = {
        "status": "ok",
        "agent_initialized": agent_wrapper is not None,
    }

    if agent_wrapper is not None:
        health_status["mcp_connected"] = hasattr(agent_wrapper.agent, "session") and agent_wrapper.agent.session is not None
        health_status["agent_name"] = agent_wrapper.agent.agent.name
        health_status["model_name"] = agent_wrapper.agent.args.get("model_name", "unknown")

    return health_status


@app.get("/test-mcp-connection")
async def test_mcp_connection():
    """Test the connection to the MCP server"""
    global mcp_session, exit_stack

    try:
        # If we already have a session, test it
        if mcp_session:
            try:
                tools_response = await mcp_session.list_tools()
                return {
                    "status": "ok",
                    "message": "Successfully connected to MCP server",
                    "tools": [tool.name for tool in tools_response.tools]
                }
            except Exception:
                # If error, close existing session and create a new one
                await exit_stack.aclose()
                mcp_session = None

        # Create a new MCP session - connect to running server
        server_params = StdioServerParameters(
            command="python",
            args=["-m", "mindsdb", "--api=mcp"],
            env=None
        )

        stdio_transport = await exit_stack.enter_async_context(stdio_client(server_params))
        stdio, write = stdio_transport
        session = await exit_stack.enter_async_context(ClientSession(stdio, write))

        await session.initialize()

        # Save the session for future use
        mcp_session = session

        # Get available tools
        tools_response = await session.list_tools()

        return {
            "status": "ok",
            "message": "Successfully connected to MCP server",
            "tools": [tool.name for tool in tools_response.tools]
        }
    except Exception as e:
        logger.error(f"Error connecting to MCP server: {str(e)}")
        error_detail = f"Error connecting to MCP server: {str(e)}. Make sure MindsDB server is running with MCP enabled: python -m mindsdb --api=mysql,mcp,http"
        raise HTTPException(status_code=500, detail=error_detail)


async def init_agent(agent_name: str, project_name: str, mcp_host: str, mcp_port: int):
    """Initialize the agent"""
    global agent_wrapper

    try:
        logger.info(f"Initializing MCP agent '{agent_name}' in project '{project_name}'")
        logger.info(f"Connecting to MCP server at {mcp_host}:{mcp_port}")
        logger.info("Make sure MindsDB server is running with MCP enabled: python -m mindsdb --api=mysql,mcp,http")

        agent_wrapper = create_mcp_agent(
            agent_name=agent_name,
            project_name=project_name,
            mcp_host=mcp_host,
            mcp_port=mcp_port
        )

        logger.info("Agent initialized successfully")
        return True
    except Exception as e:
        logger.error(f"Failed to initialize agent: {str(e)}")
        return False


@app.on_event("shutdown")
async def shutdown_event():
    """Clean up resources on server shutdown"""
    global agent_wrapper, exit_stack

    if agent_wrapper:
        await agent_wrapper.cleanup()

    await exit_stack.aclose()


async def run_server_async(
    agent_name: str,
    project_name: str = "mindsdb",
    mcp_host: str = "127.0.0.1",
    mcp_port: int = 47337,
    host: str = "0.0.0.0",
    port: int = 8000
):
    """Run the FastAPI server"""
    # Initialize the agent
    success = await init_agent(agent_name, project_name, mcp_host, mcp_port)
    if not success:
        logger.error("Failed to initialize agent. Make sure MindsDB server is running with MCP enabled.")
        return 1

    return 0


def run_server(
    agent_name: str,
    project_name: str = "mindsdb",
    mcp_host: str = "127.0.0.1",
    mcp_port: int = 47337,
    host: str = "0.0.0.0",
    port: int = 8000
):
    """Run the FastAPI server"""
    logger.info("Make sure MindsDB server is running with MCP enabled: python -m mindsdb --api=mysql,mcp,http")
    # Initialize database
    from mindsdb.interfaces.storage import db
    db.init()

    # Run initialization in the event loop
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    result = loop.run_until_complete(run_server_async(agent_name, project_name, mcp_host, mcp_port))
    if result != 0:
        return result
    # Run the server
    logger.info(f"Starting server on {host}:{port}")
    uvicorn.run(app, host=host, port=port)
    return 0


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run a LiteLLM-compatible API server for MCP agent")
    parser.add_argument("--agent", type=str, required=True, help="Name of the agent to use")
    parser.add_argument("--project", type=str, default="mindsdb", help="Project containing the agent")
    parser.add_argument("--mcp-host", type=str, default="127.0.0.1", help="MCP server host")
    parser.add_argument("--mcp-port", type=int, default=47337, help="MCP server port")
    parser.add_argument("--host", type=str, default="0.0.0.0", help="Host to bind the server to")
    parser.add_argument("--port", type=int, default=8000, help="Port to run the server on")

    args = parser.parse_args()

    run_server(
        agent_name=args.agent,
        project_name=args.project,
        mcp_host=args.mcp_host,
        mcp_port=args.mcp_port,
        host=args.host,
        port=args.port
    )



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __version__ as version, __description__ as description

try:
    from .huggingface_handler import HuggingFaceHandler as Handler

    import_error = None
except Exception as e:
    Handler = None
    import_error = e


title = "Hugging Face"
name = "huggingface"
type = HANDLER_TYPE.ML
icon_path = "icon.svg"
permanent = False
execution_method = "subprocess_keep"

__all__ = ["Handler", "version", "name", "type", "title", "description", "import_error", "icon_path"]
